###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.db.services.action_chain import simplify_ops
from vistrails.db.domain import DBAction, DBAdd, DBDelete, DBChange
import copy

def create_delete_op_chain(object, parent=(None, None)):
    opChain = []
    for (obj, parentType, parentId) in object.db_children(parent, 
                                                          for_action=True):
        op = DBDelete(id=-1,
                      what=obj.vtType,
                      objectId=obj.db_id,
                      parentObjType=parentType,
                      parentObjId=parentId,
                      )
        opChain.append(op)
    return opChain

def create_add_op_chain(object, parent=(None, None)):
    opChain = []
    object = copy.copy(object)
    adds = object.db_children(parent, True, for_action=True)
    adds.reverse()
    for (obj, parentType, parentId) in adds:
        op = DBAdd(id=-1,
                   what=obj.vtType,
                   objectId=obj.db_id,
                   parentObjType=parentType,
                   parentObjId=parentId,
                   data=obj,
                   )
        opChain.append(op)
    return opChain

def create_change_op_chain(old_obj, new_obj, parent=(None,None)):
    opChain = []
    new_obj = copy.copy(new_obj)
    deletes = old_obj.db_children(parent, for_action=True)
    deletes.pop()
    for (obj, parentType, parentId) in deletes:
        op = DBDelete(id=-1,
                      what=obj.vtType,
                      objectId=obj.db_id,
                      parentObjType=parentType,
                      parentObjId=parentId,
                      )
        opChain.append(op)

    adds = new_obj.db_children(parent, True, for_action=True)
    (obj, parentType, parentId) = adds.pop()
    op = DBChange(id=-1,
                  what=obj.vtType,
                  oldObjId=old_obj.db_id,
                  newObjId=obj.db_id,
                  parentObjType=parentType,
                  parentObjId=parentId,
                  data=new_obj,
                  )
    opChain.append(op)

    adds.reverse()
    for (obj, parentType, parentId) in adds:
        op = DBAdd(id=-1,
                   what=obj.vtType,
                   objectId=obj.db_id,
                   parentObjType=parentType,
                   parentObjId=parentId,
                   data=obj,
                   )
    return opChain

def create_copy_op_chain(object, parent=(None,None), id_scope=None):
    opChain = []
    id_remap = {}
    object = copy.copy(object)

    adds = object.db_children(parent, True, for_action=True)
    adds.reverse()
    for (obj, parentType, parentId) in adds:
        if parentId is not None:
            parentId = id_remap[(parentType, parentId)]
        new_id = id_scope.getNewId(obj.vtType)
        id_remap[(obj.vtType, obj.db_id)] = new_id
        obj.db_id = new_id
        op = DBAdd(id=-1,
                   what=obj.vtType,
                   objectId=obj.db_id,
                   parentObjType=parentType,
                   parentObjId=parentId,
                   data=obj,
                   )
        opChain.append(op)
    return opChain
    

def create_action(action_list):
    """create_action(action_list: list) -> DBAction
    where action_list is a list of tuples
     (
      type, 
      object, 
      new_obj=None,
      parent_type=None,
      parent_id=None,
    )
    Example: create_action([('add', module1), ('delete', connection2)])

    """
    ops = []
    for tuple in action_list:
        if tuple[0] == 'add' and len(tuple) >= 2:
            if len(tuple) >= 4:
                ops.extend(create_add_op_chain(tuple[1], (tuple[2], tuple[3])))
            else:
                ops.extend(create_add_op_chain(tuple[1]))
        elif tuple[0] == 'delete' and len(tuple) >= 2:
            if len(tuple) >= 4:
                ops.extend(create_delete_op_chain(tuple[1], 
                                                  (tuple[2], tuple[3])))
            else:
                ops.extend(create_delete_op_chain(tuple[1]))
        elif tuple[0] == 'change' and len(tuple) >= 3:
            if len(tuple) >= 5:
                ops.extend(create_change_op_chain(tuple[1], tuple[2],
                                                  (tuple[3], tuple[4])))
            else:
                ops.extend(create_change_op_chain(tuple[1], tuple[2]))
        else:
            msg = "unable to interpret action tuple " + tuple.__str__()
            raise ValueError(msg)
    action = DBAction(id=-1,
                      operations=ops)
    return action

def create_action_from_ops(ops, simplify=False):
    if simplify:
        ops = simplify_ops(ops)

    action = DBAction(id=-1,
                      operations=ops)
    return action



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
import sys
import os
import vistrails.db.services.io
from vistrails.db.domain import DBProvDocument, DBProvEntity, DBProvActivity, \
    DBProvAgent, DBProvGeneration, DBProvUsage, DBProvAssociation, \
    DBVtConnection, DBRefProvEntity, DBRefProvPlan, DBRefProvActivity, \
    DBRefProvAgent, DBIsPartOf, IdScope, DBGroupExec, DBLoopExec, DBModuleExec, \
    DBWorkflowExec, DBFunction, DBParameter, DBGroup, DBAbstraction
from vistrails.db.services.vistrail import materializeWorkflow

def create_prov_document(entities, activities, agents, connections, usages,
                         generations, associations):
    return DBProvDocument(prov_entitys=entities,
                          prov_activitys=activities,
                          prov_agents=agents,
                          vt_connections=connections,
                          prov_usages=usages,
                          prov_generations=generations,
                          prov_associations=associations)

def create_prov_entity_from_workflow(id_scope, workflow):
    return DBProvEntity(id='e' + str(id_scope.getNewId(DBProvEntity.vtType)),
                        vt_id=workflow.db_id,
                        prov_type='prov:Plan',
                        prov_label=workflow._db_name,
                        prov_value=None,
                        vt_type='vt:workflow',
                        vt_desc=None,
                        vt_package=None,
                        vt_version=workflow._db_version,
                        vt_cache=None,
                        vt_location_x=None,
                        vt_location_y=None,
                        is_part_of=None)

def create_prov_entity_from_module(id_scope, module, is_part_of):

    # getting module label defined by the user
    desc = None
    for db_annotation in module.db_annotations:
        if db_annotation._db_key == '__desc__':
            desc = db_annotation._db_value
            break
    
    return DBProvEntity(id='e' + str(id_scope.getNewId(DBProvEntity.vtType)),
                        vt_id=module._db_id,
                        prov_type='prov:Plan',
                        prov_label=module.db_name,
                        prov_value=None,
                        vt_type='vt:module',
                        vt_desc=desc,
                        vt_package=module.db_package,
                        vt_version=module.db_version,
                        vt_cache=module.db_cache,
                        vt_location_x=module.db_location._db_x,
                        vt_location_y=module.db_location._db_y,
                        is_part_of=is_part_of)
    
def create_is_part_of(prov_object):
    return DBIsPartOf(prov_ref=prov_object._db_id)
    
def create_prov_entity_from_group(id_scope, group, is_part_of):

    # getting group label defined by the user
    desc = None
    for db_annotation in group.db_annotations:
        if db_annotation._db_key == '__desc__':
            desc = db_annotation._db_value
            break
    
    return DBProvEntity(id='e' + str(id_scope.getNewId(DBProvEntity.vtType)),
                        vt_id=group.db_id,
                        prov_type='prov:Plan',
                        prov_label=group.db_name,
                        prov_value=None,
                        vt_type='vt:group',
                        vt_desc=desc,
                        vt_package=None,
                        vt_version=None,
                        vt_cache=group.db_cache,
                        vt_location_x=group.db_location._db_x,
                        vt_location_y=group.db_location._db_y,
                        is_part_of=is_part_of)
    
def create_prov_entity_from_abstraction(id_scope, abstraction, is_part_of):

    # getting abstraction label defined by the user
    desc = None
    for db_annotation in abstraction.db_annotations:
        if db_annotation._db_key == '__desc__':
            desc = db_annotation._db_value
            break
    
    return DBProvEntity(id='e' + str(id_scope.getNewId(DBProvEntity.vtType)),
                        vt_id=abstraction.db_id,
                        prov_type='prov:Plan',
                        prov_label=abstraction.db_name,
                        prov_value=None,
                        vt_type='vt:subworkflow',
                        vt_desc=desc,
                        vt_package=None,
                        vt_version=None,
                        vt_cache=abstraction.db_cache,
                        vt_location_x=abstraction.db_location._db_x,
                        vt_location_y=abstraction.db_location._db_y,
                        is_part_of=is_part_of)
    
def create_prov_entity_from_function(id_scope, function):
    values = []
    types = []
    aliases = []
    for parameter in function._db_parameters:
        values.append(str(parameter._db_val))
        types.append(str(parameter._db_type))
        if parameter._db_alias:
            aliases.append(parameter._db_alias)
        else:
            aliases.append('None')
        
    value = ''
    type = ''
    alias = None
    if len(values) == 1:
        value = values[0]
        type = types[0]
        alias = aliases[0]
    elif len(values) > 1:
        value = '(' + ','.join(values) + ')'
        type = '(' + ','.join(types) + ')'
        alias = '(' + ','.join(aliases) + ')'
    
    return DBProvEntity(id='e' + str(id_scope.getNewId(DBProvEntity.vtType)),
                        vt_id=function.db_id,
                        prov_type='vt:data',
                        prov_label=function.db_name,
                        prov_value=value,
                        vt_type=type,
                        vt_desc=alias,
                        vt_package=None,
                        vt_version=None,
                        vt_cache=None,
                        vt_location_x=None,
                        vt_location_y=None,
                        is_part_of=None)
    
def create_prov_entity_for_data(id_scope, conn):
    return DBProvEntity(id='e' + str(id_scope.getNewId(DBProvEntity.vtType)),
                        vt_id=conn.db_id,
                        prov_type='vt:data',
                        prov_label=None,
                        prov_value=None,
                        vt_type=None,
                        vt_desc=None,
                        vt_package=None,
                        vt_version=None,
                        vt_cache=None,
                        vt_location_x=None,
                        vt_location_y=None,
                        is_part_of=None)
    
def create_vt_connection(id_scope, source, dest, mapping):
    return DBVtConnection(id='c' + str(id_scope.getNewId(DBVtConnection.vtType)),
                          vt_source=mapping[source._db_moduleId]._db_id,
                          vt_dest=mapping[dest._db_moduleId]._db_id,
                          vt_source_port=source.db_name,
                          vt_dest_port=dest.db_name,
                          vt_source_signature=source._db_signature,
                          vt_dest_signature=dest._db_signature)
    
def create_prov_agent_from_user(id_scope, user):
    return DBProvAgent(id='ag' + str(id_scope.getNewId(DBProvAgent.vtType)),
                       vt_id=None,
                       prov_type='prov:Person',
                       prov_label=user,
                       vt_machine_os=None,
                       vt_machine_architecture=None,
                       vt_machine_processor=None,
                       vt_machine_ram=None)
    
def create_prov_agent_from_machine(id_scope, machine):
    return DBProvAgent(id='ag' + str(id_scope.getNewId(DBProvAgent.vtType)),
                       vt_id=machine.db_id,
                       prov_type='vt:machine',
                       prov_label=machine.db_name,
                       vt_machine_os=machine._db_os,
                       vt_machine_architecture=machine._db_architecture,
                       vt_machine_processor=machine._db_processor,
                       vt_machine_ram=machine._db_ram)
    
def create_prov_association(prov_activity, prov_agent, prov_entity):
    ref_prov_activity = DBRefProvActivity(prov_ref=prov_activity._db_id)
    ref_prov_agent = DBRefProvAgent(prov_ref=prov_agent._db_id)
    ref_prov_entity = DBRefProvPlan(prov_ref=prov_entity._db_id)
    
    return DBProvAssociation(prov_activity=ref_prov_activity,
                             prov_agent=ref_prov_agent,
                             prov_plan=ref_prov_entity,
                             prov_role=None)
    
def create_prov_activity_from_wf_exec(id_scope, wf_exec):
    return DBProvActivity(id='a' + str(id_scope.getNewId(DBProvActivity.vtType)),
                          vt_id=wf_exec._db_id,
                          startTime=wf_exec._db_ts_start,
                          endTime=wf_exec._db_ts_end,
                          vt_type='vt:wf_exec',
                          vt_cached=None,
                          vt_completed=wf_exec._db_completed,
                          vt_machine_id=None,
                          vt_error=None,
                          is_part_of=None)
    
def create_prov_activity_from_exec(id_scope, module_exec, machine_id, is_part_of):
    type = ''
    if module_exec.vtType == DBModuleExec.vtType:
        type = 'vt:module_exec'
    elif module_exec.vtType == DBGroupExec.vtType:
        type = 'vt:group_exec'
    else:
        # something is wrong...
        pass
    return DBProvActivity(id='a' + str(id_scope.getNewId(DBProvActivity.vtType)),
                          vt_id=module_exec._db_id,
                          startTime=module_exec._db_ts_start,
                          endTime=module_exec._db_ts_end,
                          vt_type=type,
                          vt_cached=module_exec._db_cached,
                          vt_completed=module_exec._db_completed,
                          vt_machine_id=machine_id,
                          vt_error=module_exec._db_error,
                          is_part_of=is_part_of)
    
def create_prov_usage(prov_activity, prov_entity):
    ref_prov_activity = DBRefProvActivity(prov_ref=prov_activity._db_id)
    ref_prov_entity = DBRefProvEntity(prov_ref=prov_entity._db_id)
    
    return DBProvUsage(prov_activity=ref_prov_activity,
                       prov_entity=ref_prov_entity,
                       prov_role=None)
    
def create_prov_generation(prov_entity, prov_activity):
    ref_prov_entity = DBRefProvEntity(prov_ref=prov_entity._db_id)
    ref_prov_activity = DBRefProvActivity(prov_ref=prov_activity._db_id)
    
    return DBProvGeneration(prov_entity=ref_prov_entity,
                            prov_activity=ref_prov_activity,
                            prov_role=None)

def create_prov(workflow, version, log):
    id_scope = IdScope()
    entities = []
    activities = []
    agents = []
    connections = []
    usages = []
    generations = []
    associations = []
    
    # mapping between VT ids and PROV objects
    entities_map = {}
    agents_map = {}
    
    # mapping between module ids and their functions
    module_functions = {}
    
    # mapping between module ids and module objects
    module_list = {}
    
    # mapping between module ids and source connections
    source_conn = {}
    
    # mapping between module ids and destination connections
    dest_conn = {}
    
    # mapping between connection ids and PROV entities for data
    prov_data_conn = {}
    
    # mapping between function ids and their PROV entities
    prov_functions = {}
    
    # mapping of machine ids to corresponding PROV agents
    machines = {}
    
    ############################################################################
    def get_modules_and_conn(prov_workflow, workflow):
        
        # modules
        for module in workflow._db_modules:
#            print "Entity name:", module.name
#            print module.id

            module_list[module._db_id] = module
            
            # group
            if module.vtType == DBGroup.vtType:
                vt_part = create_is_part_of(prov_workflow)
                prov_group = create_prov_entity_from_group(id_scope=id_scope,
                                                           group=module,
                                                           is_part_of=vt_part)
                entities_map[module._db_id] = prov_group
                entities.append(prov_group)
                get_modules_and_conn(prov_group, module.db_workflow)
            
            # abstraction (subworkflow)
            elif module.vtType == DBAbstraction.vtType:
                vt_part = create_is_part_of(prov_workflow)
                prov_abstraction = create_prov_entity_from_abstraction(id_scope=id_scope,
                                                                       abstraction=module,
                                                                       is_part_of=vt_part)
                entities_map[module._db_id] = prov_abstraction
                entities.append(prov_abstraction)
                #get_modules_and_conn(prov_abstraction, module.db_workflow)
            
            # module
            else:
                vt_part = create_is_part_of(prov_workflow)
                prov_module = create_prov_entity_from_module(id_scope=id_scope,
                                                             module=module,
                                                             is_part_of=vt_part)
                entities_map[module._db_id] = prov_module
                entities.append(prov_module)
                
            module_functions[module._db_id] = module._db_functions
            for function in module._db_functions:
                prov_data = create_prov_entity_from_function(id_scope, function)
                prov_functions[function.db_id] = prov_data
        
        # connections
        for conn in workflow._db_connections:
            # storing information about connections
            # used to create entities for input and output data
            source = conn.db_ports_type_index['source']
            dest = conn.db_ports_type_index['destination']
            if not source_conn.has_key(source._db_moduleId):
                source_conn[source._db_moduleId] = []
            if not dest_conn.has_key(dest._db_moduleId):
                dest_conn[dest._db_moduleId] = []
            source_conn[source._db_moduleId].append(conn)
            dest_conn[dest._db_moduleId].append(conn)
            
            prov_data = create_prov_entity_for_data(id_scope, conn)
            prov_data_conn[conn.db_id] = [prov_data, False]
            
            vt_connection = create_vt_connection(id_scope, source, dest, entities_map)
            connections.append(vt_connection)
            
        return True
    ############################################################################

    def get_execs(exec_, parent_exec, prov_agent):
        
        # module or group execution
        if exec_.vtType != DBLoopExec.vtType:
            
#            print "Exec name:", exec_.module_name
#            print exec_.module_id

            # machine (PROV agent)
            machine_id = None
#            prov_machine = machines[exec_.machine_id][0]
#            machine_id = prov_machine._db_id
#            if not machines[exec_.machine_id][1]:
#                agents.append(prov_machine)
#                machines[exec_.machine_id][1] = True
            
            # PROV activity
            vt_part = create_is_part_of(parent_exec)
            prov_activity = create_prov_activity_from_exec(id_scope,
                                                           exec_,
                                                           machine_id,
                                                           vt_part)
            
            try:
                functions = module_functions[exec_._db_module_id]
            except Exception:
                activities.append(prov_activity)
                return True
                
            for function in functions:
                prov_data = prov_functions[function.db_id]
                
                prov_usage = create_prov_usage(prov_activity, prov_data)
                usages.append(prov_usage)
                
            if dest_conn.has_key(exec_._db_module_id):
                connections = dest_conn[exec_._db_module_id]
                for connection in connections:
                    prov_input_data, inserted = prov_data_conn[connection.db_id]
                    if not inserted:
                        entities.append(prov_input_data)
                        prov_data_conn[connection.db_id][1] = True
                    
                    prov_usage = create_prov_usage(prov_activity, prov_input_data)
                    usages.append(prov_usage)
                
            if (prov_activity._db_vt_error is None) or (prov_activity._db_vt_error == ''):
                if source_conn.has_key(exec_._db_module_id):
                    connections = source_conn[exec_._db_module_id]
                    for connection in connections:
                        prov_output_data, inserted = prov_data_conn[connection.db_id]
                        if not inserted:
                            entities.append(prov_output_data)
                            prov_data_conn[connection.db_id][1] = True
                            
                        prov_generation = create_prov_generation(prov_output_data, prov_activity)
                        generations.append(prov_generation)
            
            activities.append(prov_activity)
            
            # PROV entity associated
            prov_module_entity = entities_map[exec_._db_module_id]
            prov_association = create_prov_association(prov_activity, prov_agent, prov_module_entity)
            associations.append(prov_association)
            
            if exec_.vtType == DBModuleExec.vtType:
                for loop_exec in exec_._db_loop_execs:
                    get_execs(loop_exec, prov_activity, prov_agent)
            elif exec_.vtType == DBGroupExec.vtType:
                for item in exec_._db_item_execs:
                    get_execs(item, prov_activity, prov_agent)
            else:
                # something is wrong...
                pass
            
        # loop execution
        elif exec_.vtType == DBLoopExec.vtType:
            # here, the parent execution is related to the ControlFlow:Map module
            # TODO: should get the input values from Map, and create Function +
            # Parameter?
            
            # there must be only one item exec per loop exec - otherwise, something
            # is wrong
            item_exec = exec_.item_execs[0]
            get_execs(item_exec, parent_exec, prov_agent)
        
        else:
            # something is wrong...
            pass
        
        return True
    ############################################################################
    
    # workflow
    prov_workflow = create_prov_entity_from_workflow(id_scope, workflow)
    entities_map[workflow.db_id] = prov_workflow
    entities.append(prov_workflow)
    
    # getting modules and connections 
    get_modules_and_conn(prov_workflow, workflow)
    
    # storing input data
    for id in prov_functions:
        entities.append(prov_functions[id])
    
    # machines
    for machine in log._db_machines:
        machines[machine.db_id] = (create_prov_agent_from_machine(id_scope, machine), False)
    
    # executions
    for exec_ in log._db_workflow_execs:
        if exec_._db_parent_version != version:
            continue
        prov_agent = None
        if exec_._db_user not in agents_map:
            prov_agent = create_prov_agent_from_user(id_scope, exec_._db_user)
            agents_map[exec_._db_user] = prov_agent
            agents.append(prov_agent)
        else:
            prov_agent = agents_map[exec_._db_user]
        
        # creating PROV activity
        prov_activity = create_prov_activity_from_wf_exec(id_scope, exec_)
        activities.append(prov_activity)
        
        # creating association with PROV entity
        prov_association = create_prov_association(prov_activity, prov_agent, prov_workflow)
        
        for item in exec_._db_item_execs:
            get_execs(item, prov_activity, prov_agent)
    
    # PROV Document
    return create_prov_document(entities=entities,
                                activities=activities,
                                agents=agents,
                                connections=connections,
                                usages=usages,
                                generations=generations,
                                associations=associations)
    
def add_group_portSpecs_index(workflow):
    basic_pkg = get_vistrails_basic_pkg_id()
    def process_group(group):
        def get_port_name(module):
            port_name = None
            for function in module._db_functions:
                if function.db_name == 'name':
                    port_name = function.db_parameters[0].db_val
            return port_name
        g_workflow = group.db_workflow
        group.db_portSpecs_name_index = {}
        for module in g_workflow.db_modules:
            if module.db_name == 'InputPort' and module.db_package == basic_pkg:
                port_name = get_port_name(module)
                # FIXME add sigstring to DBPortSpec
                group.db_portSpecs_name_index[(port_name, 'input')] = \
                    DBPortSpec(id=-1,
                               name=port_name,
                               type='input')
            elif module.db_name == 'OutputPort' and \
                    module.db_package == basic_pkg:
                port_name = get_port_name(module)
                # FIXME add sigstring to DBPortSpec
                group.db_portSpecs_name_index[(port_name, 'output')] = \
                    DBPortSpec(id=-1,
                               name=port_name,
                               type='output')
            elif module.db_name == 'Group' and module.db_package == basic_pkg:
                process_group(module)

    for module in workflow.db_modules:
        if module.vtType == DBGroup.vtType:
            process_group(module)
    
def create_prov_from_vistrail(vistrail, version, log):
    workflow = materializeWorkflow(vistrail, version)
    add_group_portSpecs_index(workflow)
    return create_prov(workflow, version, log)
    
def run(vistrail_xml, version, log_xml, output_fname):
    from vistrails.db.persistence import DAOList
    from vistrails.core.vistrail.vistrail import Vistrail
    import vistrails.db.services.io
    
    vistrail = vistrails.db.services.io.open_vistrail_from_xml(vistrail_xml)
    log = vistrails.db.services.io.open_log_from_xml(log_xml, was_appended=True)
    version_id = vistrail.db_get_actionAnnotation_by_key((Vistrail.TAG_ANNOTATION, version)).db_action_id
    prov_document = create_prov_from_vistrail(vistrail, int(version_id), log)
    dao_list = DAOList()
    tags = {'xmlns:prov': 'http://www.w3.org/ns/prov#',
            'xmlns:dcterms': 'http://purl.org/dc/terms/',
            'xmlns:vt': 'http://www.vistrails.org/registry.xsd',
            }
    dao_list.save_to_xml(prov_document, output_fname, tags)

if __name__ == '__main__':
    if len(sys.argv) < 4:
        print "Usage: python %s <vt_xml> <version> <log_xml> <out_xml>" % sys.argv[0]
        sys.exit(1)
    run(*sys.argv[1:])




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.db.domain import DBTag

import unittest
import copy
from vistrails.db.domain import IdScope
import vistrails.core

class Tag(DBTag):

    ##########################################################################
    # Constructors and copy

    def __init__(self, *args, **kwargs):
        DBTag.__init__(self, *args, **kwargs)
        
    def __copy__(self):
        return Tag.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBTag.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = Tag
        return cp

    @staticmethod
    def convert(_tag):
        _tag.__class__ = Tag
    
    ##########################################################################
    # Properties

    id = DBTag.db_id
    time = DBTag.db_id
    name = DBTag.db_name

    ##########################################################################
    # Operators
    
    def __str__(self):
        """__str__() -> str - Returns a string representation of a Tag
        object. 

        """
        rep = "<tag name=%s time=%s />"
        return  rep % (str(self.name), str(self.time))

    def __eq__(self, other):
        """ __eq__(other: Annotation) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(other) != type(self):
            return False
        return self.name == other.name

    def __ne__(self, other):
        return not self.__eq__(other)

################################################################################
# Testing


class TestTag(unittest.TestCase):

    def create_tag(self, id_scope=IdScope()):
        tag = Tag(id=id_scope.getNewId(Tag.vtType),
                  name='blah')
        return tag

    def test_copy(self):
        id_scope = IdScope()
        
        t1 = self.create_tag(id_scope)
        t2 = copy.copy(t1)
        self.assertEquals(t1, t2)
        self.assertEquals(t1.id, t2.id)
        t3 = t1.do_copy(True, id_scope, {})
        self.assertEquals(t1, t3)
        self.assertNotEquals(t1.id, t3.id)

    def test_serialization(self):
        import vistrails.core.db.io
        t1 = self.create_tag()
        xml_str = vistrails.core.db.io.serialize(t1)
        t2 = vistrails.core.db.io.unserialize(xml_str, Tag)
        self.assertEquals(t1, t2)
        self.assertEquals(t1.id, t2.id)



#!/usr/bin/env python

# In this example vtkClipPolyData is used to cut a polygonal model
# of a cow in half. In addition, the open clip is closed by triangulating
# the resulting complex polygons.

import vtk
from vtk.util.misc import vtkGetDataRoot
from vtk.util.colors import peacock, tomato
VTK_DATA_ROOT = vtkGetDataRoot()

# First start by reading a cow model. We also generate surface normals for
# prettier rendering.
cow = vtk.vtkBYUReader()
cow.SetGeometryFileName(VTK_DATA_ROOT + "/Data/Viewpoint/cow.g")
cowNormals = vtk.vtkPolyDataNormals()
cowNormals.SetInputConnection(cow.GetOutputPort())

# We clip with an implicit function. Here we use a plane positioned near
# the center of the cow model and oriented at an arbitrary angle.
plane = vtk.vtkPlane()
plane.SetOrigin(0.25, 0, 0)
plane.SetNormal(-1, -1, 0)

# vtkClipPolyData requires an implicit function to define what it is to
# clip with. Any implicit function, including complex boolean combinations
# can be used. Notice that we can specify the value of the implicit function
# with the SetValue method.
clipper = vtk.vtkClipPolyData()
clipper.SetInputConnection(cowNormals.GetOutputPort())
clipper.SetClipFunction(plane)
clipper.GenerateClipScalarsOn()
clipper.GenerateClippedOutputOn()
clipper.SetValue(0.5)
clipMapper = vtk.vtkPolyDataMapper()
clipMapper.SetInputConnection(clipper.GetOutputPort())
clipMapper.ScalarVisibilityOff()
backProp = vtk.vtkProperty()
backProp.SetDiffuseColor(tomato)
clipActor = vtk.vtkActor()
clipActor.SetMapper(clipMapper)
clipActor.GetProperty().SetColor(peacock)
clipActor.SetBackfaceProperty(backProp)

# Here we are cutting the cow. Cutting creates lines where the cut
# function intersects the model. (Clipping removes a portion of the
# model but the dimension of the data does not change.)
#
# The reason we are cutting is to generate a closed polygon at the
# boundary of the clipping process. The cutter generates line
# segments, the stripper then puts them together into polylines. We
# then pull a trick and define polygons using the closed line
# segements that the stripper created.
cutEdges = vtk.vtkCutter()
cutEdges.SetInputConnection(cowNormals.GetOutputPort())
cutEdges.SetCutFunction(plane)
cutEdges.GenerateCutScalarsOn()
cutEdges.SetValue(0, 0.5)
cutStrips = vtk.vtkStripper()
cutStrips.SetInputConnection(cutEdges.GetOutputPort())
cutStrips.Update()
cutPoly = vtk.vtkPolyData()
cutPoly.SetPoints(cutStrips.GetOutput().GetPoints())
cutPoly.SetPolys(cutStrips.GetOutput().GetLines())

# Triangle filter is robust enough to ignore the duplicate point at
# the beginning and end of the polygons and triangulate them.
cutTriangles = vtk.vtkTriangleFilter()
cutTriangles.SetInput(cutPoly)
cutMapper = vtk.vtkPolyDataMapper()
cutMapper.SetInput(cutPoly)
cutMapper.SetInputConnection(cutTriangles.GetOutputPort())
cutActor = vtk.vtkActor()
cutActor.SetMapper(cutMapper)
cutActor.GetProperty().SetColor(peacock)

# The clipped part of the cow is rendered wireframe.
restMapper = vtk.vtkPolyDataMapper()
restMapper.SetInput(clipper.GetClippedOutput())
restMapper.ScalarVisibilityOff()
restActor = vtk.vtkActor()
restActor.SetMapper(restMapper)
restActor.GetProperty().SetRepresentationToWireframe()

# Create graphics stuff
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(clipActor)
ren.AddActor(cutActor)
ren.AddActor(restActor)
ren.SetBackground(1, 1, 1)
ren.ResetCamera()
ren.GetActiveCamera().Azimuth(30)
ren.GetActiveCamera().Elevation(30)
ren.GetActiveCamera().Dolly(1.5)
ren.ResetCameraClippingRange()

renWin.SetSize(300, 300)
iren.Initialize()

# Lets you move the cut plane back and forth by invoking the function
# Cut with the appropriate plane value (essentially a distance from
# the original plane).  This is not used in this code but should give
# you an idea of how to define a function to do this.
def Cut(v):
    clipper.SetValue(v)
    cutEdges.SetValue(0, v)
    cutStrips.Update()
    cutPoly.SetPoints(cutStrips.GetOutput().GetPoints())
    cutPoly.SetPolys(cutStrips.GetOutput().GetLines())
    cutMapper.Update()
    renWin.Render()
 
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from itertools import izip
import operator

from vistrails.core.data_structures.bijectivedict import Bidict
from vistrails.core.modules.utils import create_port_spec_string, parse_port_spec_string
from vistrails.core.system import get_vistrails_basic_pkg_id, \
    get_module_registry
from vistrails.core.utils import enum, VistrailsInternalError
from vistrails.core.vistrail.port_spec_item import PortSpecItem
from vistrails.db.domain import DBPortSpec, IdScope

from ast import literal_eval
import unittest
import copy

PortEndPoint = enum('PortEndPoint',
                    ['Invalid', 'Source', 'Destination'])

################################################################################

class PortSpec(DBPortSpec):

    port_type_map = Bidict([('input', 'destination'),
                            ('output', 'source'),
                            ('invalid', 'invalid')])
    end_point_map = Bidict([('source', PortEndPoint.Source),
                            ('destination', PortEndPoint.Destination),
                            ('invalid', PortEndPoint.Invalid)])

    ##########################################################################
    # Constructors and copy

    def __init__(self, *args, **kwargs):
        signature = None
        if 'signature' in kwargs:
            signature = kwargs['signature']
            del kwargs['signature']
        sigstring = None
        if 'sigstring' in kwargs:
            sigstring = kwargs['sigstring']
            del kwargs['sigstring']
        defaults = None
        if 'defaults' in kwargs:
            defaults = kwargs['defaults']
            del kwargs['defaults']
        labels = None
        if 'labels' in kwargs:
            labels = kwargs['labels']
            del kwargs['labels']
        values = None
        if 'values' in kwargs:
            values = kwargs['values']
            del kwargs['values']
        entry_types = None
        if 'entry_types' in kwargs:
            entry_types = kwargs['entry_types']
            del kwargs['entry_types']

        if 'items' in kwargs and 'portSpecItems' not in kwargs:
            kwargs['portSpecItems'] = kwargs['items']
            del kwargs['items']

        if 'optional' not in kwargs:
            kwargs['optional'] = 0 # False
        elif not isinstance(kwargs['optional'], (int, long)):
            if isinstance(kwargs['optional'], bool):
                if kwargs['optional']:
                    kwargs['optional'] = 1
                else:
                    kwargs['optional'] = 0
            else:
                raise VistrailsInternalError("Cannot parse 'optional' kw "
                                             "-- must be an int or bool")
        if 'min_conns' not in kwargs:
            kwargs['min_conns'] = 0
        elif kwargs['optional'] == 1 and kwargs['min_conns'] > 0:
            raise VistrailsInternalError("A mandatory port cannot be set "
                                         "to optional")
        if 'max_conns' not in kwargs:
            kwargs['max_conns'] = -1
        if kwargs['min_conns'] >= 0 and kwargs['max_conns'] >= 0 and \
                kwargs['min_conns'] > kwargs['max_conns']:
            raise VistrailsInternalError("Minimum number of connections "
                                         "cannot be greater than maximum "
                                         "number of connections")
            
        if 'sort_key' not in kwargs:
            kwargs['sort_key'] = -1
        if 'depth' not in kwargs:
            kwargs['depth'] = 0
        if 'id' not in kwargs:
            kwargs['id'] = -1
        if 'tooltip' in kwargs:
            self._tooltip = kwargs['tooltip']
            del kwargs['tooltip']
        else:
            self._tooltip = None

        if 'docstring' in kwargs:
            self._docstring = kwargs['docstring']
            del kwargs['docstring']
        else:
            self._docstring = None
        if 'shape' in kwargs:
            self._shape = kwargs['shape']
            del kwargs['shape']
        else:
            self._shape = None

        DBPortSpec.__init__(self, *args, **kwargs)

        if sum(1 for container in (self.port_spec_items, signature, sigstring)
               if container) > 1:
            raise ValueError("Please specify only one of portSpecItems,"
                             " signature, or sigstring kwargs.")
        self.create_spec_items(self.port_spec_items, signature, sigstring, 
                               defaults, labels, values, entry_types)

        self._short_sigstring = None
        # if signature is not None:
        #     self.create_entries(signature)
        # if not self.sigstring and self._entries is not None:
        #     # create sigstring from entries
        #     self.create_sigstring_and_descriptors()
# DAKOOP: removed this---we will check in module_registry and pipeline 
# validation, this way, we can let errors go all the way up
#         elif self._entries is None and self.sigstring:
#             # create entries from sigstring
#             self.create_entries_and_descriptors()
#         else:
#             raise VistrailsInternalError("Need to specify signature or "
#                                          "sigstring to create PortSpec")
        # if self._entries is not None and self._tooltip is None:
        #     self.create_tooltip()
        self.is_valid = True

    def __copy__(self):
        return PortSpec.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBPortSpec.do_copy(self, new_ids, id_scope, id_remap)
        cp._short_sigstring = self._short_sigstring
        cp._tooltip = self._tooltip
        cp._shape = self._shape
        cp._docstring = self._docstring
        cp.is_valid = self.is_valid
        cp.__class__ = PortSpec
        # if cp._entries is not None:
        #     cp.create_tooltip()
        return cp

    @staticmethod
    def convert(_port_spec):
        if _port_spec.__class__ == PortSpec:
            return
        _port_spec.__class__ = PortSpec
        for _port_spec_item in _port_spec.db_portSpecItems:
            PortSpecItem.convert(_port_spec_item)
        _port_spec._short_sigstring = None
        _port_spec._tooltip = None
        _port_spec._shape = None
        _port_spec._docstring = None
        _port_spec.is_valid = True

        _port_spec.port_spec_items.sort(key=operator.attrgetter('db_pos'))

    @staticmethod
    def from_sigstring(sigstring):
        """from_sig(sigstring: string) -> PortSpec

        Returns a portspec from the given sigstring.

        """
        return PortSpec(sigstring=sigstring)

    ##########################################################################
    # Properties

    id = DBPortSpec.db_id
    name = DBPortSpec.db_name
    type = DBPortSpec.db_type
    optional = DBPortSpec.db_optional
    sort_key = DBPortSpec.db_sort_key
    min_conns = DBPortSpec.db_min_conns
    max_conns = DBPortSpec.db_max_conns
    _depth = DBPortSpec.db_depth
    port_spec_items = DBPortSpec.db_portSpecItems
    items = DBPortSpec.db_portSpecItems

    def _get_sigstring(self):
        return create_port_spec_string([i.spec_tuple 
                                        for i in self.port_spec_items])
    sigstring = property(_get_sigstring)

    def is_mandatory(self):
        return (self.min_conns > 0)

    def _get_labels(self):
        return [i.label for i in self.port_spec_items]
    labels = property(_get_labels)

    def _get_defaults(self):
        return [i.default for i in self.port_spec_items]
    defaults = property(_get_defaults)
    
    def _get_short_sigstring(self):
        if self._short_sigstring is None:
            self.create_tooltip()
        return self._short_sigstring
    short_sigstring = property(_get_short_sigstring)

    def _get_signature(self):
        signature = []
        for i in self.port_spec_items:
            signature.append((i.descriptor.module, i.label))
        return signature
    signature = property(_get_signature)

    def _get_depth(self):
        return self._depth or 0
    def _set_depth(self, depth):
        self._depth = depth
    depth = property(_get_depth, _set_depth)

    def toolTip(self):
        if self._tooltip is None:
            self.create_tooltip()
        return self._tooltip

    def shape(self):
        return self._shape
    
    def docstring(self):
        return self._docstring

    def descriptors(self):
        return [i.descriptor for i in self.port_spec_items]

    ##########################################################################
    # Methods

    def _resize_attrs(self, target, *lists):
        for rlist in lists:
            if len(target) > len(rlist):
                rlist.extend(None for i in xrange(len(target)-len(rlist)))

    def _set_attrs(self, item, *attrs):
        attr_order = ['default', 'label', 'values', 'entry_type']
        if item is None:
            kwargs = dict(izip(attr_order, attrs))
            return kwargs
        else:
            for (attr_key, attr) in izip(attr_order, attrs):
                if attr is not None:
                    setattr(item, attr_key, attr)


    def create_spec_items(self, items=None, signature=None, sigstring=None, 
                          defaults=None, labels=None, values=None, 
                          entry_types=None):
        if defaults is None:
            defaults = []
        elif isinstance(defaults, basestring):
            defaults = literal_eval(defaults)
        if labels is None:
            labels = []
        elif isinstance(labels, basestring):
            labels = literal_eval(labels)
        if values is None:
            values = []
        elif isinstance(values, basestring):
            values = literal_eval(values)
        if entry_types is None:
            entry_types = []
        elif isinstance(entry_types, basestring):
            entry_types = literal_eval(entry_types)
        attrs = [defaults, labels, values, entry_types]
        if items:
            self.set_items(items, *attrs)
        elif signature is not None:
            items = self.get_items_from_entries(signature, *attrs)
        elif sigstring is not None:
            items = self.get_items_from_sigstring(sigstring, *attrs)
        self.port_spec_items = items

    def set_items(self, items, *attrs):
        self._resize_attrs(items, *attrs)
        for i, item_tuple in enumerate(izip(items, *attrs)):
            item_tuple[0].pos = i
            self._set_attrs(*item_tuple)

    def get_items_from_entries(self, signature, *attrs):
        # This is reasonably messy code. The intent is that a
        # signature given by the user in a call like this
        # add_input_port(module, name, signature) should be one of the
        # following:

        # type only: add_input_port(_, _, Float)
        # type plus description: add_input_port(_, _, (Float, 'radius'))

        # multiple parameters, where each parameter can be either of the above:
        # add_input_port(_, _, [Float, (Integer, 'count')])

        registry = get_module_registry()
        entries = []
        def canonicalize(sig_item):
            if isinstance(sig_item, tuple):
                # assert len(sig_item) == 2
                # assert isinstance(sig_item[0], type)
                # assert isinstance(sig_item[1], str)
                descriptor = registry.get_descriptor(sig_item[0])
                label = sig_item[1]
                return (descriptor, label)
            elif isinstance(sig_item, list):
                descriptor = registry.get_descriptor_by_name(
                    get_vistrails_basic_pkg_id(), 'List')
                return (descriptor, None)
            else:
                # isinstance(sig_item, type):
                return (registry.get_descriptor(sig_item), None)

        # def _add_entry(sig_item):
        ps_items = []
        if not isinstance(signature, list):
            signature = [signature]
        self._resize_attrs(signature, *attrs)
        for i, item_tuple in enumerate(izip(signature, *attrs)):
            descriptor, item_label = canonicalize(item_tuple[0])
            kwargs = self._set_attrs(None, *item_tuple[1:])
            if not kwargs['label']:
                if item_label != "<no description>":
                    kwargs['label'] = item_label
            ps_item = PortSpecItem(pos=i, 
                                   package=descriptor.identifier,
                                   module=descriptor.name,
                                   namespace=descriptor.namespace,
                                   **kwargs)
            ps_items.append(ps_item)
        return ps_items

    def get_items_from_sigstring(self, sigstring, *attrs):
        ps_items = []
        specs_list = parse_port_spec_string(sigstring)
        if len(specs_list) == 0:
            return ps_items

        self._resize_attrs(specs_list, *attrs)
        for i, item_tuple in enumerate(izip(specs_list, *attrs)):
            kwargs = self._set_attrs(None, *item_tuple[1:])
            ps_item = PortSpecItem(pos=i,
                                   package=item_tuple[0][0],
                                   module=item_tuple[0][1],
                                   namespace=item_tuple[0][2],
                                   **kwargs)
            ps_items.append(ps_item)
        return ps_items

    def create_tooltip(self):
        """Creates a short_sigstring that does not include package names for
        use with the tooltip. Note, however, that such sigstrings
        can't be used to reconstruct a spec. They should only be used
        for human-readable purposes.
        """

        self._short_sigstring = \
            "(" + ",".join(d.name for d in self.descriptors()) + ")"
        if self.type in ['input', 'output']:
            port_string = self.type.capitalize()
        else:
            port_string = 'Invalid'
        _depth = " (depth %s)" % self.depth if self.depth else ''
        self._tooltip = "%s port %s\n%s%s" % (port_string,
                                            self.name,
                                            self._short_sigstring,
                                            _depth)
        
    ##########################################################################
    # Operators
    
    def __str__(self):
        """__str__() -> str - Returns a string representation of an PortSpec
        object. 

        """
        rep = "<portSpec id=%s name=%s type=%s signature=%s depth=%s />"
        return  rep % (str(self.id), str(self.name), 
                       str(self.type), str(self.sigstring), str(self.depth))

    def __eq__(self, other):
        """ __eq__(other: PortSpec) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if self is None and other is None:
            return True
        if type(self) != type(other) or \
                self.name != other.name or \
                self.type != other.type:
            return False
        if len(self.descriptors()) != len(other.descriptors()):
            return False
        for (mine, their) in izip(self.descriptors(), other.descriptors()):
            if mine != their:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def type_equals(self, other):
        """type_equals(other: PortSpec) -> Bool

        Checks equality ignoring description strings. Only cares about types.
        Does not do subtyping or supertyping: match must be perfect.

        """
        if self is None and other is None:
            return True
        if len(self.descriptors()) != len(other.descriptors()):
            return False
        for (mine, their) in izip(self.descriptors(), other.descriptors()):
            if mine != their:
                return False
        return True
        
    def key_no_id(self):
        """key_no_id(): tuple. returns a tuple that identifies
        the port without caring about ids. Used for sorting
        port lists."""
        return (self.type,
                self.name,
                self.signature)

################################################################################
# Testing


class TestPortSpec(unittest.TestCase):

    def create_port_spec(self, id_scope=IdScope()):
        # FIXME add a valid port spec
        port_spec = PortSpec(id=id_scope.getNewId(PortSpec.vtType),
                             name='SetValue',
                             type='input',
                             sigstring='(%s:String)' % \
                                 get_vistrails_basic_pkg_id(),
                             )
        return port_spec

    def test_copy(self):
        id_scope = IdScope()
        
        s1 = self.create_port_spec(id_scope)
        s2 = copy.copy(s1)
        self.assertEquals(s1, s2)
        self.assertEquals(s1.id, s2.id)
        s3 = s1.do_copy(True, id_scope, {})
        self.assertEquals(s1, s3)
        self.assertNotEquals(s1.id, s3.id)

    def test_serialization(self):
        import vistrails.core.db.io
        s1 = self.create_port_spec()
        xml_str = vistrails.core.db.io.serialize(s1)
        s2 = vistrails.core.db.io.unserialize(xml_str, PortSpec)
        self.assertEquals(s1, s2)
        self.assertEquals(s1.id, s2.id)

    def test_create_from_signature(self):
        from vistrails.core.modules.basic_modules import Float
        port_spec = PortSpec(id=-1,
                             name="SetXYZ",
                             type='input',
                             signature=[(Float, "x"), (Float, "y"), 
                                        (Float, "z")])

    def test_create_from_items(self):
        basic_pkg = get_vistrails_basic_pkg_id()
        item_a = PortSpecItem(pos=0,
                              package=basic_pkg,
                              module="Integer",
                              label="a",
                              default="123")
        item_b = PortSpecItem(pos=1,
                              package=basic_pkg,
                              module="String",
                              label="b",
                              default="abc")
        port_spec = PortSpec(id=-1,
                             name="SetValue",
                             type='input',
                             portSpecItems=[item_a, item_b])
        



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from vistrails.core.modules.config import ModuleSettings
from vistrails.core.modules.vistrails_module import Module

import numpy as np
from sklearn.base import ClassifierMixin
from sklearn import datasets
from sklearn.cross_validation import train_test_split, cross_val_score
from sklearn.metrics import SCORERS, roc_curve
from sklearn.grid_search import GridSearchCV as _GridSearchCV
from sklearn.pipeline import make_pipeline
from sklearn.utils.testing import all_estimators


def try_convert(input_string):
    if not isinstance(input_string, basestring):
        # already converted
        return input_string
    if input_string.isdigit():
        return int(input_string)
    try:
        return float(input_string)
    except ValueError:
        return input_string

# backport of odd estimators that we don't want to include
dont_test = ['SparseCoder', 'EllipticEnvelope', 'DictVectorizer',
             'LabelBinarizer', 'LabelEncoder', 'MultiLabelBinarizer',
             'TfidfTransformer', 'IsotonicRegression', 'OneHotEncoder',
             'RandomTreesEmbedding', 'FeatureHasher', 'DummyClassifier',
             'DummyRegressor', 'TruncatedSVD', 'PolynomialFeatures']


###############################################################################
# Example datasets

class Digits(Module):
    """Example dataset: digits.
    """
    _settings = ModuleSettings(namespace="datasets")
    _output_ports = [("data", "basic:List", {'shape': 'circle'}),
                     ("target", "basic:List", {'shape': 'circle'})]

    def compute(self):
        data = datasets.load_digits()
        self.set_output("data", data.data)
        self.set_output("target", data.target)


class Iris(Module):
    """Example dataset: iris.
    """
    _settings = ModuleSettings(namespace="datasets")
    _output_ports = [("data", "basic:List", {'shape': 'circle'}),
                     ("target", "basic:List", {'shape': 'circle'})]

    def compute(self):
        data = datasets.load_iris()
        self.set_output("data", data.data)
        self.set_output("target", data.target)


###############################################################################
# Base classes

class Estimator(Module):
    """Base class for all sklearn estimators.
    """
    _settings = ModuleSettings(abstract=True)
    _output_ports = [("model", "Estimator", {'shape': 'diamond'})]


class SupervisedEstimator(Estimator):
    """Base class for all sklearn classifier modules."""
    _settings = ModuleSettings(abstract=True)

    def compute(self):
        # get parameters, try to convert strings to float / int
        params = dict([(p, try_convert(self.get_input(p))) for p in self.inputPorts
                       if p not in ["training_data", "training_target"]])
        clf = self._estimator_class(**params)
        if "training_data" in self.inputPorts:
            training_data = np.vstack(self.get_input("training_data"))
            training_target = self.get_input("training_target")
            clf.fit(training_data, training_target)
        self.set_output("model", clf)


class UnsupervisedEstimator(Estimator):
    """Base class for all sklearn transformer modules."""
    _settings = ModuleSettings(abstract=True)

    def compute(self):
        params = dict([(p, try_convert(self.get_input(p))) for p in self.inputPorts
                       if p not in ["training_data", "training_target"]])
        trans = self._estimator_class(**params)
        if "training_data" in self.inputPorts:
            training_data = np.vstack(self.get_input("training_data"))
            trans.fit(training_data)
        self.set_output("model", trans)


class Predict(Module):
    """Apply a learned scikit-learn classifier model to test data.
    """
    # TODO : data depth=1
    _input_ports = [("model", "Estimator", {'shape': 'diamond'}),
                    ("data", "basic:List", {'shape': 'circle'})]
    _output_ports = [("prediction", "basic:List", {'shape': 'circle'}),
                     ("decision_function", "basic:List")]

    def compute(self):
        clf = self.get_input("model")
        data = self.get_input("data")
        predictions = clf.predict(data)
        decision_function = clf.decision_function(data)
        self.set_output("prediction", predictions)
        self.set_output("decision_function", decision_function)


class Transform(Module):
    """Apply a learned scikit-learn transformer to test data.
    """
    _input_ports = [("model", "Estimator", {'shape': 'diamond'}),
                    ("data", "basic:List", {'shape': 'circle'})]
    _output_ports = [("transformed_data", "basic:List", {'shape': 'circle'})]

    def compute(self):
        trans = self.get_input("model")
        data = self.get_input("data")
        transformed_data = trans.transform(data)
        self.set_output("transformed_data", transformed_data)


###############################################################################
# Cross-validation


class TrainTestSplit(Module):
    """Split data into training and testing randomly."""
    _settings = ModuleSettings(namespace="cross-validation")
    _input_ports = [("data", "basic:List", {'shape': 'circle'}),
                    ("target", "basic:List", {'shape': 'circle'}),
                    ("test_size", "basic:Float", {"defaults": [.25]})]
    _output_ports = [("training_data", "basic:List", {'shape': 'circle'}),
                     ("training_target", "basic:List", {'shape': 'circle'}),
                     ("test_data", "basic:List", {'shape': 'circle'}),
                     ("test_target", "basic:List", {'shape': 'circle'})]

    def compute(self):
        X_train, X_test, y_train, y_test = \
            train_test_split(self.get_input("data"), self.get_input("target"),
                             test_size=try_convert(self.get_input("test_size")))
        self.set_output("training_data", X_train)
        self.set_output("training_target", y_train)
        self.set_output("test_data", X_test)
        self.set_output("test_target", y_test)


class CrossValScore(Module):
    """Split data into training and testing randomly."""
    _settings = ModuleSettings(namespace="cross-validation")
    _input_ports = [("model", "Estimator", {'shape': 'diamond'}),
                    ("data", "basic:List", {'shape': 'circle'}),
                    ("target", "basic:List", {'shape': 'circle'}),
                    ("metric", "basic:String", {"defaults": ["accuracy"]}),
                    ("folds", "basic:Integer", {"defaults": ["3"]})]
    _output_ports = [("scores", "basic:List")]

    def compute(self):
        model = self.get_input("model")
        data = self.get_input("data")
        target = self.get_input("target")
        metric = self.get_input("metric")
        folds = self.get_input("folds")
        scores = cross_val_score(model, data, target, scoring=metric, cv=folds)
        self.set_output("scores", scores)

###############################################################################
# Meta Estimators


class GridSearchCV(Estimator):
    """Perform cross-validated grid-search over a parameter grid."""
    _input_ports = [("model", "Estimator", {'shape': 'diamond'}),
                    ("parameters", "basic:Dictionary"),
                    ("data", "basic:List", {'shape': 'circle'}),
                    ("target", "basic:List", {'shape': 'circle'}),
                    ("metric", "basic:String", {"defaults": ["accuracy"]}),
                    ("folds", "basic:Integer", {"defaults": ["3"]})]
    _output_ports = [("scores", "basic:List"), ("model", "Estimator", {'shape': 'diamond'}),
                     ("best_parameters", "basic:Dictionary"),
                     ("best_score", "basic:Float")]

    def compute(self):
        base_model = self.get_input("model")
        grid = _GridSearchCV(base_model,
                             param_grid=self.get_input("parameters"),
                             cv=self.get_input("folds"),
                             scoring=self.get_input("metric"))
        if "data" in self.inputPorts:
            data = np.vstack(self.get_input("data"))
            target = self.get_input("target")
            grid.fit(data, target)
            self.set_output("scores", grid.grid_scores_)
            self.set_output("best_parameters", grid.best_params_)
            self.set_output("best_score", grid.best_score_)
        self.set_output("model", grid)


class Pipeline(Estimator):
    """Chain estimators to form a pipeline."""
    _input_ports = [("training_data", "basic:List", {'shape': 'circle'}),
                    ("training_target", "basic:List", {'shape': 'circle'}),("model1", "Estimator", {'shape': 'diamond'}),
                    ("model2", "Estimator", {'optional': True, 'shape': 'diamond'}),
                    ("model3", "Estimator", {'optional': True, 'shape': 'diamond'}),
                    ("model4", "Estimator", {'optional': True, 'shape': 'diamond'})]

    def compute(self):
        models = ["model%d" % d for d in range(1, 5)]
        steps = [self.get_input(model) for model in models if model in self.inputPorts]
        pipeline = make_pipeline(*steps)
        if "training_data" in self.inputPorts:
            training_data = np.vstack(self.get_input("training_data"))
            training_target = self.get_input("training_target")
            pipeline.fit(training_data, training_target)
        self.set_output("model", pipeline)

###############################################################################
# Metrics


class Score(Module):
    """Compute a model performance metric."""
    _settings = ModuleSettings()
    _input_ports = [("model", "Estimator", {'shape': 'diamond'}),
                    ("data", "basic:List", {'shape': 'circle'}),
                    ("target", "basic:List", {'shape': 'circle'}),
                    ("metric", "basic:String", {"defaults": ["accuracy"]})]
    _output_ports = [("score", "basic:Float")]

    def compute(self):
        scorer = SCORERS[self.get_input("metric")]
        score = scorer(self.get_input("model"), self.get_input("data"), self.get_input("target"))
        self.set_output("score", score)


class ROCCurve(Module):
    """Compute a ROC curve."""
    _settings = ModuleSettings(namespace="metrics")
    _input_ports = [("model", "Estimator", {'shape': 'diamond'}),
                    ("data", "basic:List", {'shape': 'circle'}),
                    ("target", "basic:List", {'shape': 'circle'})]
    _output_ports = [("fpr", "basic:List"),
                     ("tpr", "basic:List")]

    def compute(self):
        model = self.get_input("model")
        data = self.get_input("data")
        if hasattr(model, "decision_function"):
            dec = model.decision_function(data)
        else:
            dec = model.predict_proba(data)[:, 1]
        fpr, tpr, _ = roc_curve(self.get_input("target"), dec)
        self.set_output("fpr", fpr)
        self.set_output("tpr", tpr)


###############################################################################<F2>
# Classifiers and Regressors

def make_module(name, Estimator, namespace, supervised=False, Base=None):
    input_ports = [("training_data", "basic:List", {'shape': 'circle'})]
    if supervised:
        input_ports.append(("training_target", "basic:List", {'shape': 'circle'}))
    est = Estimator()
    input_ports.extend([(param, "basic:String", {'optional': True}) for param
                        in est.get_params()])
    _settings = ModuleSettings(namespace=namespace)
    if Base is None:
        if supervised:
            Base = SupervisedEstimator
        else:
            Base = UnsupervisedEstimator
    new_class = type(name, (Base,),
                     {'_input_ports': input_ports, '_settings': _settings,
                      '_estimator_class': Estimator, '__doc__':
                      Estimator.__doc__})
    return new_class


def discover_supervised():
    classifiers = all_estimators(type_filter="classifier")
    regressors = all_estimators(type_filter="regressor")
    classes = []
    for name, Est in classifiers + regressors:
        if issubclass(Est, ClassifierMixin):
            namespace = "classifiers"
        else:
            namespace = "regressors"
        classes.append(make_module(name, Est, namespace, supervised=True))
    return classes


###############################################################################<F2>
# Clustering

def discover_clustering():
    return [make_module(name, Est, "clustering") for (name, Est) in
            all_estimators(type_filter="cluster")]


###############################################################################
# Transformers

def discover_unsupervised_transformers():
    # also: random tree embedding
    transformers = all_estimators(type_filter="transformer")
    classes = []
    for name, Est in transformers:
        if name in dont_test:
            continue
        module = Est.__module__.split(".")[1]
        if module not in ['decomposition', 'kernel_approximation',
                          'neural_network', 'preprocessing', 'random_projection']:
            # the manifold module does not really provide transformers and is
            # handled elsewhere (there is usually no transform method).
            continue
        classes.append(make_module(name, Est, namespace=module))
    return classes


def discover_feature_selection():
    # also: random tree embedding
    transformers = all_estimators(type_filter="transformer")
    classes = []
    for name, Est in transformers:
        if name in dont_test:
            continue
        module = Est.__module__.split(".")[1]
        if module != "feature_selection" or name == "GenericUnivariateSelect":
            continue
        classes.append(make_module(name, Est, namespace=module, supervised=True))
    return classes


###############################################################################
# Manifold learning
# unfortunately this has a sightly different interface due to the fact that
# most manifold algorithms can't generalize to new data.
class ManifoldLearner(Module):
    """Base class for all sklearn manifold modules.
    """
    _settings = ModuleSettings(abstract=True)
    _output_ports = [("transformed_data", "basic:List", {'shape': 'circle'})]

    def compute(self):
        params = dict([(p, try_convert(self.get_input(p))) for p in self.inputPorts
                       if p not in ["training_data"]])
        trans = self._estimator_class(**params)
        training_data = np.vstack(self.get_input("training_data"))
        transformed_data = trans.fit_transform(training_data)
        self.set_output("transformed_data", transformed_data)


def discover_manifold_learning():
    # also: random tree embedding
    transformers = all_estimators()
    classes = []
    for name, Est in transformers:
        if name in dont_test:
            continue
        module = Est.__module__.split(".")[1]
        if module != "manifold":
            continue
        classes.append(make_module(name, Est, namespace=module, Base=ManifoldLearner))
    return classes


_modules = [Digits, Iris, Estimator, SupervisedEstimator,
            UnsupervisedEstimator, ManifoldLearner, Predict, Transform,
            TrainTestSplit, Score, ROCCurve, CrossValScore, GridSearchCV,
            Pipeline]
_modules.extend(discover_supervised())
_modules.extend(discover_clustering())
_modules.extend(discover_unsupervised_transformers())
_modules.extend(discover_feature_selection())
_modules.extend(discover_manifold_learning())



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""Hasher class for vistrail items."""
from __future__ import division

from vistrails.core.cache.utils import hash_list

try:
    import hashlib
    sha_hash = hashlib.sha1
except ImportError:
    import sha
    sha_hash = sha.new

##############################################################################

class Hasher(object):

    @staticmethod
    def parameter_signature(p, constant_hasher_map={}):
        k = (p.identifier, p.type, p.namespace)
        custom_hasher = constant_hasher_map.get(k, None)
        if custom_hasher:
            return custom_hasher(p)
        else:
            hasher = sha_hash()
            u = hasher.update
            u(p.type)
            u(p.identifier)
            u(p.namespace or "")
            u(p.strValue)
            u(p.name)
            u(p.evaluatedStrValue)
            return hasher.digest()

    @staticmethod
    def function_signature(function, constant_hasher_map={}):
        hasher = sha_hash()
        u = hasher.update
        u(function.name)
        u(function.returnType)
        u(hash_list(function.params,
                    Hasher.parameter_signature,
                    constant_hasher_map))
        return hasher.digest()

    @staticmethod
    def control_param_signature(control_param, constant_hasher_map={}):
        hasher = sha_hash()
        u = hasher.update
        u(control_param.name)
        u(control_param.value)
        return hasher.digest()

    @staticmethod
    def connection_signature(c):
        hasher = sha_hash()
        u = hasher.update
        u(c.source.name)
        u(c.destination.name)
        return hasher.digest()

    @staticmethod
    def connection_subpipeline_signature(c, source_sig, dest_sig):
        """Returns the signature for the connection, including source and dest
        subpipelines

        """
        hasher = sha_hash()
        u = hasher.update
        u(Hasher.connection_signature(c))
        u(source_sig)
        u(dest_sig)
        return hasher.digest()

    @staticmethod
    def module_signature(obj, constant_hasher_map={}):
        hasher = sha_hash()
        u = hasher.update
        u(obj.module_descriptor.name)
        u(obj.module_descriptor.package)
        u(obj.module_descriptor.namespace or '')
        u(obj.module_descriptor.package_version or '')
        u(obj.module_descriptor.version or '')
        u(hash_list(obj.functions, Hasher.function_signature,
                    constant_hasher_map))
        u(hash_list(obj.control_parameters, Hasher.control_param_signature,
                    constant_hasher_map))
        return hasher.digest()

    @staticmethod
    def subpipeline_signature(module_sig, upstream_sigs):
        """Returns the signature for a subpipeline, given the signatures for
        the upstream pipelines and connections.

        WARNING: For efficiency, upstream_sigs is mutated!

        """
        hasher = sha_hash()
        hasher.update(module_sig)
        upstream_sigs.sort()
        for pipeline_connection_sig in upstream_sigs:
            hasher.update(pipeline_connection_sig)
        return hasher.digest()

    @staticmethod
    def compound_signature(sig_list):
        """compound_signature(list of signatures) -> sha digest returns the
        signature of the compound object formed by the list of
        signatures, assuming the list order is irrelevant

        """
        hasher = sha_hash()
        for h in sorted(sig_list):
            hasher.update(h)
        return hasher.digest()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file provides the basic infrastructure for extensible parameter
exploration. This allows user-defined constants to be used as dimensions
in parameter exploration, provided the user implements the appropriate
API in the classes.
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.modules.basic_modules import Color
from vistrails.core import debug
from vistrails.core.modules.paramexplore import IntegerLinearInterpolator, \
   FloatLinearInterpolator, RGBColorInterpolator, HSVColorInterpolator

from vistrails.gui.common_widgets import QStringEdit
from vistrails.gui.modules.constant_configuration import ColorChooserButton
from vistrails.gui.modules.python_source_configure import PythonEditor
from vistrails.gui.modules.utils import get_param_explore_widget_list
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.utils import show_warning
from vistrails.core.utils import all, unimplemented

##############################################################################

class QParameterEditor(QtGui.QWidget):
    """
    QParameterEditor specifies the method used for interpolating
    parameter values. It suppports Linear Interpolation, List and
    User-define function.
    
    """
    def __init__(self, param_info, size, parent=None):
        """ QParameterEditor(param_info: ParameterInfo,
                             size: int, parent=None: QWidget) -> QParameterEditor
        Put a stacked widget and a popup button
        
        """
        QtGui.QWidget.__init__(self, parent)
        self._param_info = param_info
        self.type = param_info.spec.descriptor.name
        self.defaultValue = param_info.value
        
        hLayout = QtGui.QHBoxLayout(self)
        hLayout.setMargin(0)
        hLayout.setSpacing(0)
        self.setLayout(hLayout)

        descriptor = param_info.spec.descriptor

        self.stackedEditors = QtGui.QStackedWidget()
        self.stackedEditors.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                          QtGui.QSizePolicy.Maximum)
        self._exploration_widgets = []

        def add_exploration_widget(wd):
            self._exploration_widgets.append(wd)
            self.stackedEditors.addWidget(wd)

        for widget_class in get_param_explore_widget_list(descriptor):
            new_widget = widget_class(param_info, size)
            add_exploration_widget(new_widget)

        add_exploration_widget(QListInterpolationEditor(param_info, size))
        add_exploration_widget(QUserFunctionEditor(param_info, size))

        hLayout.addWidget(self.stackedEditors)

        self.selector = QParameterEditorSelector(param_info, self._exploration_widgets)
        self.connect(self.selector.actionGroup,
                     QtCore.SIGNAL('triggered(QAction*)'),
                     self.changeInterpolator)
        hLayout.addWidget(self.selector)
        self.selector.initAction()

    def changeInterpolator(self, action):
        """ changeInterpolator(action: QAction) -> None        
        Bring the correct interpolation editing widget to front in the
        stacked widget
        
        """
        widgetIdx = action.data()
        if widgetIdx < self.stackedEditors.count():
            self.stackedEditors.setCurrentIndex(widgetIdx)

    def selectInterpolator(self, type):
        """ selectInterpolator(type: string) -> None
        Programatically (without requiring a user action) selects
        the interpolator specified by the name 'type'.  If no
        matching 'type' is found, no action is taken.
        
        """
        types = [widget.exploration_name for widget in self._exploration_widgets]
        if type in types:
            type_idx = types.index(type)
            self.selector._actions[type_idx].trigger()

class QParameterEditorSelector(QtGui.QToolButton):
    """
    QParameterEditorSelector is a button with a down arrow allowing
    users to select which type of interpolator he/she wants to use
    
    """
    def __init__(self, param_info, widget_list, parent=None):
        """ QParameterEditorSelector(param_info: ParameterInfo,
        widget_list: list of widgets that conform to the parameter
        exploration widget interface.
        """
        QtGui.QToolButton.__init__(self, parent)
        self._param_info = param_info
        self.type = param_info.spec.descriptor.name
        self.setAutoRaise(True)
        self.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.setPopupMode(QtGui.QToolButton.InstantPopup)
        
        self.setText(unichr(0x25bc)) # Down triangle

        self.actionGroup = QtGui.QActionGroup(self)

        self._actions = []
        for widget in widget_list:
            action = QtGui.QAction(widget.exploration_name, self.actionGroup)
            self._actions.append(action)

        aId = 0
        for action in self.actionGroup.actions():
            action.setCheckable(True)
            action.setData(aId)
            aId += 1

        menu = QtGui.QMenu(self)
        menu.addActions(self.actionGroup.actions())
        self.setMenu(menu)

    def initAction(self):
        """ initAction() -> None
        Select the first choice of selector based on self.type
        
        """
        self._actions[0].trigger()

##############################################################################

class BasePEWidget(object):
    def get_value(self):
        unimplemented()
    def set_value(self):
        unimplemented()
        
class QIntegerLineEdit(QtGui.QLineEdit, BasePEWidget):
    def __init__(self, param_info, parent=None):
        QtGui.QLineEdit.__init__(self, param_info.value, parent)
        self.setValidator(QtGui.QIntValidator(self))
    def get_value(self):
        return int(str(self.text()))
    def set_value(self, str_value):
        self.setText(str_value)
        
class QFloatLineEdit(QtGui.QLineEdit, BasePEWidget):
    def __init__(self, param_info, parent=None):
        QtGui.QLineEdit.__init__(self, param_info.value, parent)
        self.setValidator(QtGui.QDoubleValidator(self))
    def get_value(self):
        return float(str(self.text()))
    def set_value(self, str_value):
        self.setText(str_value)
##############################################################################

def make_interpolator(widget_class, interpolator_class, name):
    class InterpolationEditor(QtGui.QWidget):
        """
        QLinearInterpolationEditor is the actual widget allowing users to
        edit his/her linear interpolation parameters.

        """
        def __init__(self, param_info, size, parent=None):
            QtGui.QWidget.__init__(self, parent)
            self._param_info = param_info
            self.type = param_info.spec.descriptor.name

            hLayout = QtGui.QHBoxLayout(self)
            hLayout.setMargin(0)
            hLayout.setSpacing(0)
            self.setLayout(hLayout)

            self.fromEdit = widget_class(param_info)
            hLayout.addWidget(self.fromEdit)

            hLayout.addSpacing(5)

            rightArrow = QtGui.QLabel()
            pixmap = self.style().standardPixmap(QtGui.QStyle.SP_ArrowRight)
            rightArrow.setPixmap(CurrentTheme.RIGHT_ARROW_PIXMAP)
            hLayout.addWidget(rightArrow)

            hLayout.addSpacing(5)

            self.toEdit = widget_class(param_info)
            hLayout.addWidget(self.toEdit)
            self.exploration_name = name

        def get_values(self, size):
            """ get_values(size: int) -> tuple
            Return the interpolated list containing 'size' values

            """
            begin = self.fromEdit.get_value()
            end = self.toEdit.get_value()
            lerp = interpolator_class(begin, end, size)
            return lerp.get_values()
    return InterpolationEditor

FloatExploreWidget = make_interpolator(QFloatLineEdit,
                                       FloatLinearInterpolator,
                                       'Linear Interpolation')
IntegerExploreWidget = make_interpolator(QIntegerLineEdit,
                                         IntegerLinearInterpolator,
                                         'Linear Interpolation')

class PEColorChooserButton(ColorChooserButton, BasePEWidget):

    def __init__(self, param_info, parent=None):
        ColorChooserButton.__init__(self, parent)
        try:
            r,g,b = [int(float(i) * 255) for i in param_info.value.split(',')]
        except ValueError:
            r,g,b = (0.0, 0.0, 0.0)
        self.setColor(QtGui.QColor(r,g,b))
        self.setFixedHeight(22)
        self.setSizePolicy(QtGui.QSizePolicy.Expanding,
                           QtGui.QSizePolicy.Fixed)

    def get_value(self):
        return Color.to_string(self.qcolor.redF(),
                               self.qcolor.greenF(),
                               self.qcolor.blueF())
        
    def set_value(self, str_value):
        color = str_value.split(',')
        qcolor = QtGui.QColor(float(color[0])*255,
                              float(color[1])*255,
                              float(color[2])*255)
        self.setColor(qcolor)

RGBExploreWidget = make_interpolator(PEColorChooserButton,
                                     RGBColorInterpolator,
                                     'RGB Interpolation')
HSVExploreWidget = make_interpolator(PEColorChooserButton,
                                     HSVColorInterpolator,
                                     'HSV Interpolation')

##############################################################################

class QListInterpolationEditor(QtGui.QWidget):
    """
    QListInterpolationEditor is the actual widget allowing users to
    enter a list of values for interpolation
    
    """
    def __init__(self, param_info, size, parent=None):
        """ QListInterpolationEditor(param_info: ParameterInfo, parent: QWidget)
                                     -> QListInterpolationEditor
        Construct an edit box with a button for bringing up the dialog
        
        """
        QtGui.QWidget.__init__(self, parent)
        self._param_info = param_info
        self.type = param_info.spec.descriptor.name
        
        hLayout = QtGui.QHBoxLayout(self)
        hLayout.setMargin(0)
        hLayout.setSpacing(0)
        self.setLayout(hLayout)
        
        self.listValues = QtGui.QLineEdit()
        if self.type=='String':
            self.listValues.setText("['%s']" % param_info.value.replace("'", "\'"))
        else:
            self.listValues.setText('[%s]' % param_info.value)
        self.listValues.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                      QtGui.QSizePolicy.Maximum)
        self.listValues.home(False)
        hLayout.addWidget(self.listValues)

        self.connect(self.listValues, QtCore.SIGNAL('textEdited(QString)'),
                     self.values_were_edited)

        inputButton = QtGui.QToolButton()
        inputButton.setText('...')
        self.connect(inputButton, QtCore.SIGNAL('clicked()'),
                     self.editListValues)
        hLayout.addWidget(inputButton)
        self.exploration_name = 'List'
        self._str_values = [param_info.value]

    def values_were_edited(self, new_text):
        """values_were_edited(new_text): None

        Connected to self.listValues.textEdited. Updates self._str_values.
        
        NB: Allowing the user to edit the LineEdit field directly is
        not a very good idea, because we don't know what are the
        syntactic rules for the translate_to_python() calls in
        arbitrary classes.  Right now, I'm assuming removing the
        leading and trailing brackets and splitting on ',' is
        enough. (in passing, The previous call to eval() is just
        broken is a general scenario like we have now)

        For example, this will break horribly if the user manually edits
        a list of strings with commas in them."""

        t = str(new_text)
        if len(t) < 2:
            self._str_values = []
            return
        if not (t[0] == '[' and t[-1] == ']'):
            self._str_values = []
        else: 
            self._str_values = t[1:-1].split(',')
            if self._param_info.spec.descriptor.name=='String':
                for i, val in enumerate(self._str_values):
                    val = val.strip()
                    if len(val) >= 2 and  \
                            ((val[0] == "'" and val[-1] == "'") or 
                             (val[0] == '"' and val[-1] == '"')):
                        self._str_values[i] = val.strip()[1:-1]

    def get_values(self, count):
        """ get_values(count) -> []        
        Convert the list values into a list

        count should be an integer with the expected size of the list (given
        by the dimension 'size' in the exploration)
        
        """

        param_info = self._param_info
        module = param_info.spec.descriptor.module
        result = [module.translate_to_python(m)
                  for m in self._str_values]
        if len(result) != count:
            show_warning('Inconsistent Size',
                         'One of the <i>%s</i>\'s list '
                         'interpolated '
                         'values has a different '
                         'size from the step count. '
                         'Expected %d, got %d instead. '
                         'Parameter Exploration aborted.'
                         % (self.type, count, len(result)))
            return None
        return result

    def editListValues(self):
        """ editListValues() -> None
        Show a dialog for editing the values
        
        """
        dialog = QListEditDialog(self.type, self._str_values, None)
        if dialog.exec_() == QtGui.QDialog.Accepted:
            values = dialog.getList()
            self._str_values = [str(v) for v in values]
            if self.type == 'String':
                values = ["'%s'" % v.replace("'", "\'")
                          for v in values]
            self.listValues.setText('[%s]' % ', '.join(values))
            self.listValues.home(False)
        dialog.deleteLater()

##############################################################################

class QListEditDialog(QtGui.QDialog):
    """
    QListEditDialog provides an interface for user to edit a list of
    values and export to a string
    
    """
    def __init__(self, pType, strValues, parent=None):
        """ QListEditDialog(pType: str, strValues: list, parent: QWidget)
                            -> QListEditDialog
        Parse values and setup the table
        
        """
        QtGui.QDialog.__init__(self, parent)
        self.pType = pType
        vLayout = QtGui.QVBoxLayout()
        vLayout.setMargin(0)
        vLayout.setSpacing(0)
        self.setLayout(vLayout)
        
        label = QtGui.QLabel("Please enter values in boxes below. Drag "
                             "rows up and down to arrange your list values. "
                             "'Add' appends an empty value to the list. "
                             "And 'Del' removes the selected values.")
        label.setMargin(5)
        label.setWordWrap(True)
        vLayout.addWidget(label)

        self.table = QtGui.QTableWidget(0, 1, parent)
        self.table.setHorizontalHeaderLabels(['Values'])
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.verticalHeader().setMovable(True)
        self.table.verticalHeader().setResizeMode(
            QtGui.QHeaderView.ResizeToContents)
        self.delegate = QListEditItemDelegate()
        self.table.setItemDelegate(self.delegate)
        self.table.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
        for v in strValues:
            self.addRow(v)
        self.connect(self.table.verticalHeader(),
                     QtCore.SIGNAL('sectionMoved(int,int,int)'),
                     self.rowMoved)
        vLayout.addWidget(self.table)

        hLayout = QtGui.QHBoxLayout()        
        vLayout.addLayout(hLayout)

        okButton = QtGui.QPushButton('&OK')
        okButton.setSizePolicy(QtGui.QSizePolicy.Maximum,
                               QtGui.QSizePolicy.Maximum)
        self.connect(okButton, QtCore.SIGNAL('clicked()'), self.okButtonPressed)
        hLayout.addWidget(okButton)

        cancelButton = QtGui.QPushButton('&Cancel')
        cancelButton.setSizePolicy(QtGui.QSizePolicy.Maximum,
                                   QtGui.QSizePolicy.Maximum)
        self.connect(cancelButton, QtCore.SIGNAL('clicked()'), self.reject)
        hLayout.addWidget(cancelButton)

        addButton = QtGui.QPushButton('&Add')
        addButton.setIcon(CurrentTheme.ADD_STRING_ICON)
        addButton.setSizePolicy(QtGui.QSizePolicy.Maximum,
                                QtGui.QSizePolicy.Maximum)
        self.connect(addButton, QtCore.SIGNAL('clicked()'), self.addRow)
        hLayout.addWidget(addButton)
        
        removeButton = QtGui.QPushButton('&Del')
        removeButton.setIcon(QtGui.QIcon(
            self.style().standardPixmap(QtGui.QStyle.SP_DialogCancelButton)))
        removeButton.setSizePolicy(QtGui.QSizePolicy.Maximum,
                                   QtGui.QSizePolicy.Maximum)
        self.connect(removeButton, QtCore.SIGNAL('clicked()'),
                     self.removeSelection)
        hLayout.addWidget(removeButton)
        
    def sizeHint(self):
        """ sizeHint() -> QSize
        Return the recommended size for the widget
        
        """
        return QtCore.QSize(256, 384)

    def okButtonPressed(self):
        """ okButtonPressed() -> None
        Make sure to commit the editor data before accepting
        
        """
        self.table.itemDelegate().finishEditing()
        self.accept()

    def getList(self):
        """ getList() -> list of str values
        Return a list of values
        
        """
        result = []
        for i in xrange(self.table.rowCount()):
            logicalIndex = self.table.verticalHeader().logicalIndex(i)
            value = self.table.item(logicalIndex, 0).text()            
            result.append(str(value))
        return result

    def rowMoved(self, row, old, new):
        """ rowMove(row: int, old: int, new: int) -> None
        Renumber the vertical header labels when row moved
        
        """
        vHeader = self.table.verticalHeader()
        labels = []        
        for i in xrange(self.table.rowCount()):
            labels.append(str(vHeader.visualIndex(i)+1))
        self.table.setVerticalHeaderLabels(labels)

    def addRow(self, text=None):
        """ addRow(text: str) -> QListStringEdit
        Add an extra row to the end of the table
        
        """
        self.table.setRowCount(self.table.rowCount()+1)
        if text:
            item = QtGui.QTableWidgetItem(text)
        else:
            item = QtGui.QTableWidgetItem()
        row = self.table.rowCount()-1
        self.table.setItem(row, 0, item)

    def removeSelection(self):
        """ removeSelection() -> None
        Remove selected rows on the table
        
        """
        for item in self.table.selectedItems():
            self.table.removeRow(item.row())

##############################################################################

class QListEditItemDelegate(QtGui.QItemDelegate):
    """
    QListEditItemDelegate sets up the editor for the QListEditDialog
    table
    
    """

    def __init__(self, parent=None):
        """ QListEditItemDelegate(parent: QWidget) -> QListEditItemDelegate
        Store the uncommit editor for commit later
        
        """
        QtGui.QItemDelegate.__init__(self, parent)
        self.editor = None
        
    def createEditor(self, parent, option, index):
        """ createEditor(parent: QWidget,
                         option: QStyleOptionViewItem,
                         index: QModelIndex) -> QStringEdit
        Return the editor widget for the index
        
        """
        self.editor = QStringEdit(parent)
        return self.editor

    def setEditorData(self, editor, index):
        """ setEditorData(editor: QWidget, index: QModelIndex) -> None
        Set the editor to reflects data at index
        
        """
        editor.setText(index.data())
        editor.selectAll()

    def updateEditorGeometry(self, editor, option, index):
        """ updateEditorGeometry(editor: QStringEdit,
                                 option: QStyleOptionViewItem,
                                 index: QModelIndex) -> None
        Update the geometry of the editor based on the style option
        
        """
        editor.setGeometry(option.rect)

    def setModelData(self, editor, model, index):
        """ setModelData(editor: QStringEdit,
                         model: QAbstractItemModel,
                         index: QModelIndex) -> None
        Set the text of the editor back to the item model
        
        """
        model.setData(index, editor.text())
        self.editor = None

    def finishEditing(self):
        if self.editor:
            self.emit(QtCore.SIGNAL('commitData(QWidget*)'), self.editor)

##############################################################################

class QUserFunctionEditor(QtGui.QFrame):
    """
    QUserFunctionEditor shows user-defined interpolation function
    
    """
    def __init__(self, param_info, size, parent=None):
        """ QUserFunctionEditor(param_info: ParameterInfo, parent: QWidget)
                                -> QUserFunctionEditor
        Create a read-only line edit widget and a button for
        customizing the user-defined function
        
        """
        QtGui.QFrame.__init__(self, parent)
        self.setFrameStyle(QtGui.QFrame.Box | QtGui.QFrame.Sunken)
        self.size = -1
        self._param_info = param_info
        self.type = param_info.spec.descriptor.name
        self.defaultValue = param_info.value
        self.function = self.defaultFunction()
        
        hLayout = QtGui.QHBoxLayout(self)
        hLayout.setMargin(0)
        hLayout.setSpacing(0)
        self.setLayout(hLayout)
        
        hLayout.addSpacing(2)
        self.label = QtGui.QLabel()
        hLayout.addWidget(self.label)

        self.listValues = QtGui.QLineEdit()
        self.listValues.setFrame(False)        
        self.listValues.palette().setBrush(QtGui.QPalette.Base,
                                           QtGui.QBrush(QtCore.Qt.NoBrush))
        self.listValues.setReadOnly(True)
        self.listValues.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                      QtGui.QSizePolicy.Maximum)
        self.listValues.home(False)
        hLayout.addWidget(self.listValues)

        self.size_was_updated(size)

        inputButton = QtGui.QToolButton()
        inputButton.setText('...')
        self.connect(inputButton, QtCore.SIGNAL('clicked()'),
                     self.editFunction)
        hLayout.addWidget(inputButton)
        self.exploration_name = 'User-defined Function'

    def defaultFunction(self):
        """ defaultFunction() -> str
        Return the default function definition
        
        """
        if self.type=='String':
            quote = '"'
        else:
            quote = ''
        return 'def value(i):\n    """ value(i: int) -> value\n'\
               '    Returns the i-th value\n'\
               '    i is from 0 to <step count>-1.\n'\
               '    If this function has an error, the value will be\n'\
               '    the default constant value\n'\
               '    """\n'\
               '    # Define your function here\n'

    def get_values(self, count):
        """ get_values() -> []        
        Convert the user define function into a list. Size specifies the size
        request.
        
        """
        param_info = self._param_info
        module = param_info.spec.descriptor.module
        def get():
            values = []
            d = {}
            try:
                exec(self.function) in {}, d
            except Exception, e:
                return [module.default_value] * count
            def evaluate(i):
                try:
                    v = d['value'](i)
                    if v is None:
                        return module.default_value
                    return v
                except Exception, e:
                    debug.unexpected_exception(e)
                    return debug.format_exception(e)
            return [evaluate(i) for i in xrange(self.size)]
        result = get()
        
        if not all(module.validate(x) for x in result):
            show_warning('Failed Validation',
                         'One of the <i>%s</i>\'s user defined '
                         'functions has failed validation, '
                         'which usually means it generated a '
                         'value of a type different '
                         'than that specified by the '
                         'parameter. Parameter Exploration '
                         'aborted.' % param_info.spec.descriptor.name)
            return None
        return result
        

    def getValuesString(self):
        """ getValuesString() -> str
        Return a string representation of the parameter list
        
        """
        r = self.get_values(self.size)
        if r is None:
            return '{ERROR}'
        else:
            return '{%s}' % ','.join([str(v) for v in r])

    def editFunction(self):
        """ editFunction() -> None
        Pop up a dialog for editing user-defined function
        
        """
        dialog = QUserFunctionDialog(self.function)        
        if dialog.exec_()==QtGui.QDialog.Accepted:
            self.function = str(dialog.editor.toPlainText())
            self.listValues.setText(self.getValuesString())
        dialog.deleteLater()

    def size_was_updated(self, size):
        """ size_was_updated(size: int) -> None

        This is called whenever the size of the interpolation is changed.
        Values are re-calculated."""
        if size!=self.size:
            self.size = size
            htmlText = '<html><big>&fnof;</big>(n) <b>:</b> ' \
                       '[0,%d) &rarr; </html>' % size
            self.label.setText(htmlText)
            self.listValues.setText(self.getValuesString())

##############################################################################

class QUserFunctionDialog(QtGui.QDialog):
    """
    QUserFunctionDialog provides an interface for user to edit a
    python function
    
    """
    def __init__(self, function, parent=None):
        """ QUserFunctionDialog(function: str, parent: QWidget)
                                -> QUserFunctionDialog
        Set up a python source editor
        
        """
        QtGui.QDialog.__init__(self, parent)
        vLayout = QtGui.QVBoxLayout()
        vLayout.setMargin(0)
        vLayout.setSpacing(0)
        self.setLayout(vLayout)
        self.setWindowTitle('User-defined Function')
        
        label = QtGui.QLabel("Please define your function below. This "
                             "'value(i)' function will be iteratively called "
                             "for <step count> numbers. For each step, "
                             "it should return a value of parameter type.")
        label.setMargin(5)
        label.setWordWrap(True)
        vLayout.addWidget(label)

        self.editor = PythonEditor(self)
        self.editor.setPlainText(function)
        vLayout.addWidget(self.editor)

        hLayout = QtGui.QHBoxLayout()        
        vLayout.addLayout(hLayout)

        okButton = QtGui.QPushButton('&OK')
        okButton.setSizePolicy(QtGui.QSizePolicy.Maximum,
                               QtGui.QSizePolicy.Maximum)
        self.connect(okButton, QtCore.SIGNAL('clicked()'), self.accept)
        hLayout.addWidget(okButton)

        cancelButton = QtGui.QPushButton('&Cancel')
        cancelButton.setSizePolicy(QtGui.QSizePolicy.Maximum,
                                   QtGui.QSizePolicy.Maximum)
        self.connect(cancelButton, QtCore.SIGNAL('clicked()'), self.reject)
        hLayout.addWidget(cancelButton)
        
    def sizeHint(self):
        """ sizeHint() -> QSize
        Return the recommended size for the widget
        
        """
        return QtCore.QSize(512, 512)



#!/usr/bin/env python
# pragma: no testimport
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""auto-generates code given specs"""

# requires mako python package (easy_install Mako) and 
# uses emacs via subprocess call for python indentation
# the emacs call is slow because it checks all of the indentation

from __future__ import division

from mako.template import Template

import os
import re
import shutil
import subprocess
import sys
import tempfile
import getopt
from parser import AutoGenParser
import xml_gen_objects
import sql_gen_objects

BASE_DIR = os.path.dirname(os.getcwd())

DOMAIN_INIT = """from auto_gen import *"""
PERSISTENCE_INIT = \
"""from xml.auto_gen import XMLDAOListBase
from sql.auto_gen import SQLDAOListBase

class DAOList(dict):
    def __init__(self):
        self['xml'] = XMLDAOListBase()
        self['sql'] = SQLDAOListBase()

"""
COPYRIGHT_NOTICE = \
"""###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""

def preprocess_template(in_fname, out_fname=None):
    in_file = open(in_fname)
    if out_fname is not None:
        out_file = open(out_fname, 'w')
    else:
        out_file = sys.stdout

    single_slash = re.compile(r'\s*\\$')
    double_slash = re.compile(r'\s*\\\\$')
    triple_slash = re.compile(r'\s*\\\\\\$')
    slash_bang = re.compile(r'\s*\\!$')
    python_block = re.compile(r'^\s+<%')
    strip_whitespace = False
    for line in in_file:
        if strip_whitespace:
            line = line.lstrip()
        if python_block.search(line):
            line = python_block.sub(r'<%', line)
        if triple_slash.search(line):
            line = triple_slash.sub(r' ${"\\\\"}', line)
            strip_whitespace = False
        elif double_slash.search(line):
            line = double_slash.sub(r'\\', line)
            strip_whitespace = False
        elif slash_bang.search(line):
            line = line[:-3].rstrip()
            strip_whitespace = True
        elif single_slash.search(line):
            line = line[:-2].rstrip() + ' '
            strip_whitespace = True
        else:
            strip_whitespace = False
        out_file.write(line)

    if out_fname is not None:
        out_file.close()
    in_file.close()

def indent_python(fname):
    subprocess.Popen(["emacs", "-batch", fname, "-f", "mark-whole-buffer",
                      "-f", "indent-region", "-f", "save-buffer", "-kill"],
                     stdout=subprocess.PIPE).communicate()

def run_template(template_fname, objects, version, version_string, output_file,
                 indent=False):
    [prefix, suffix] = os.path.basename(template_fname).split('.', 1)
    (fd, p_fname) = tempfile.mkstemp(prefix=prefix, suffix=suffix)
    os.close(fd)
    try:
        preprocess_template(template_fname, p_fname)
        template = Template(filename=p_fname, module_directory='/tmp/mako')

        f = open(output_file, 'w')
        f.write(template.render(objs=objects,
                                version=version,
                                version_string=version_string))
        f.close()
        if indent:
            indent_python(output_file)
    finally:
        os.remove(p_fname)

def usage(usageDict):
    usageStr = ''
    unrequired = ''
    required = ''
    for (opt, info) in usageDict.iteritems():
        if info[1]:
            required += '-%s <%s> ' % (opt[0], info[2])
            usageStr += '    -%s <%s>  ' % (opt[0], info[2])
        else:
            if len(opt) > 1:
                unrequired += '[-%s <%s>] ' % (opt[0], info[2])
                usageStr += '    -%s <%s>  ' % (opt[0], info[2])
            else:
                unrequired += '[-%s] ' % opt[0]
                usageStr += '    -%s            ' % opt[0]
        usageStr += '%s\n' % info[0]
    print 'Usage: python generate.py %s%s\n%s' % (required, 
                                                             unrequired, 
                                                             usageStr)

def dirStructure(baseDir):
    dirs = {}
    dirs['base'] = baseDir
    dirs['specs'] = os.path.join(dirs['base'], 'specs')
    dirs['persistence'] = os.path.join(dirs['base'], 'persistence')
    dirs['domain'] = os.path.join(dirs['base'], 'domain')
    dirs['schemas'] = os.path.join(dirs['base'], 'schemas')
    dirs['xmlPersistence'] = os.path.join(dirs['persistence'], 'xml')
    dirs['sqlPersistence'] = os.path.join(dirs['persistence'], 'sql')
    dirs['xmlSchema'] = os.path.join(dirs['schemas'], 'xml')
    dirs['sqlSchema'] = os.path.join(dirs['schemas'], 'sql')
    return dirs

def makeAllDirs(dirs):
    for (name, dir) in dirs.iteritems():
        if not os.path.exists(dir):
            print "creating directory '%s'" % dir
            os.makedirs(dir)
        if name not in set(['specs', 'schemas', 'xmlSchema', 'sqlSchema']):
            init_file = os.path.join(dir, '__init__.py')
            if not os.path.exists(init_file):
                print "creating file '%s'" % init_file
                f = open(init_file, 'w')
                f.write(COPYRIGHT_NOTICE)
                if name == 'domain':
                    f.write(DOMAIN_INIT)
                elif name == 'persistence':
                    f.write(PERSISTENCE_INIT)
                else:
                    f.write('pass')
                f.close()
                    
def main(argv=None):
    options = {}
    objects = None

    optionsUsage = {'a': ('generate all database information (-p -s -x)', 
                          False),
                    'b:': ('base directory', False, 'dir'),
                    'd:': ('versions directory', False, 'dir'),
                    'p': ('generate python domain classes', False),
                    's': ('generate sql schema and persistence classes', False),
                    'x': ('generate xml schema and persistence classes', False),
                    'v:': ('vistrail version tag', True, 'version'),
                    'm': ('make all directories', False),
                    'n': ('do not change current version', False)}

    optStr = ''.join(optionsUsage.keys())
    optKeys = optStr.replace(':','')
    for idx in xrange(len(optKeys)):
        options[optKeys[idx]] = False

    try:
        (optlist, args) = getopt.getopt(sys.argv[1:], optStr)
        for opt in optlist:
            if opt[1] is not None and opt[1] != '':
                options[opt[0][1:]] = opt[1]
            else:
                options[opt[0][1:]] = True
    except getopt.GetoptError:
        usage(optionsUsage)
        return

    if options['b']:
        baseDir = options['b']
    else:
        baseDir = BASE_DIR
    baseDirs = dirStructure(baseDir)

    if not options['v']:
        usage(optionsUsage)
        return

    version = options['v']
    # make sure version use dot-style
    assert(len(version.split('.'))==3)
    versionName = 'v' + version.replace('.', '_')
    if options['d']:
        versionsDir = options['d']
    else:
        versionsDir = os.path.join(baseDir, 'versions')
    versionDir = os.path.join(versionsDir, versionName)
    versionDirs = dirStructure(versionDir)

    print baseDirs
    print versionDirs

    if options['m']:
        makeAllDirs(baseDirs)
        makeAllDirs(versionDirs)

    # check whether we should use existing specs for version
    use_base_specs = True
    if options['n'] and os.path.exists(versionDirs['specs']):
        for file in os.listdir(versionDirs['specs']):
            if file.endswith('.xml'):
                # assume we've already copied the specs
                use_base_specs = False

    if use_base_specs:
        # copy specs to version        
        print "copying base specs to version directory..."
        for file in os.listdir(baseDirs['specs']):
            if file.endswith('.xml'):
                print 'copying %s' % file
                filename = os.path.join(baseDirs['specs'], file)
                toFile = os.path.join(versionDirs['specs'], file)
                shutil.copyfile(filename, toFile)
    else:
        print "using existing specs from version directory..."

    if options['p'] or options['a']:
        # generate python domain objects
        print "generating python domain objects..."
        if objects is None:
            parser = AutoGenParser()
            objects = parser.parse(versionDirs['specs'])
        run_template('templates/domain.py.mako', objects, version, versionName,
                     os.path.join(versionDirs['domain'], 'auto_gen.py'),
                     True)

        if not options['n']:
            domainFile = os.path.join(baseDirs['domain'], '__init__.py')
            f = open(domainFile, 'w')
            f.write(COPYRIGHT_NOTICE)
            f.write('from vistrails.db.versions.%s.domain import *\n' % \
                        versionName)
            f.close()

    if options['x'] or options['a']:
        # generate xml schema and dao objects
        print "generating xml schema and dao objects..."
        if objects is None:
            parser = AutoGenParser()
            objects = parser.parse(versionDirs['specs'])
        xml_objects = xml_gen_objects.convert(objects)
        
        run_template('templates/xml_schema.xsd.mako', 
                     xml_gen_objects.convert_schema_order(xml_objects, 
                                                          'vistrail'),
                     version, versionName,
                     os.path.join(versionDirs['xmlSchema'], 'vistrail.xsd'),
                     False)

        run_template('templates/xml_schema.xsd.mako', 
                     xml_gen_objects.convert_schema_order(xml_objects, 
                                                          'workflow'),
                     version, versionName,
                     os.path.join(versionDirs['xmlSchema'], 'workflow.xsd'),
                     False)

        run_template('templates/xml_schema.xsd.mako', 
                     xml_gen_objects.convert_schema_order(xml_objects, 
                                                          'log'),
                     version, versionName,
                     os.path.join(versionDirs['xmlSchema'], 'log.xsd'),
                     False)

        run_template('templates/xml.py.mako', xml_objects, version, versionName,
                     os.path.join(versionDirs['xmlPersistence'], 'auto_gen.py'),
                     True)

    if options['s'] or options['a']:
        # generate sql schema and dao objects
        print "generating sql schema and dao objects..."
        if objects is None:
            parser = AutoGenParser()
            objects = parser.parse(versionDirs['specs'])
        sql_objects = sql_gen_objects.convert(objects)

        run_template('templates/sql_schema.sql.mako', sql_objects, 
                     version, versionName,
                     os.path.join(versionDirs['sqlSchema'], 'vistrails.sql'),
                     False)

        run_template('templates/sql_delete.sql.mako', sql_objects, 
                     version, versionName,
                     os.path.join(versionDirs['sqlSchema'], 
                                  'vistrails_drop.sql'),
                     False)
        
        run_template('templates/sql.py.mako', sql_objects,
                     version, versionName,
                     os.path.join(versionDirs['sqlPersistence'], 'auto_gen.py'),
                     True)

    if not options['n']:
        domainFile = os.path.join(baseDirs['persistence'], '__init__.py')
        f = open(domainFile, 'w')
        f.write(COPYRIGHT_NOTICE)
        f.write('from vistrails.db.versions.%s.persistence import *\n' % \
                    versionName)
        f.close()
            
if __name__ == '__main__':
    main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import os
import shutil
import sys

vistrails_src = None
if not vistrails_src:
    vistrails_src = os.path.dirname(
        os.path.dirname(os.path.dirname(sys.path[0])))
if vistrails_src not in sys.path:
    sys.path.append(vistrails_src)

import vistrails.db.services.io

from identifiers import identifier as persistence_pkg, \
    old_identifiers as persistence_old_ids
persistence_pkg_ids = set(persistence_old_ids)
persistence_pkg_ids.add(persistence_pkg)
    
def find_files(filename, version=None):
    save_bundle, save_dir = \
        vistrails.db.services.io.open_vistrail_bundle_from_zip_xml(filename)
    vistrail = save_bundle.vistrail
    # FIXME hack for now, should change in the future
    log_fname = vistrail.db_log_filename
    log = vistrails.db.services.io.open_log_from_xml(log_fname, True)

    if version:
        if isinstance(version, basestring):
            # need to lookup version number
            if version in vistrail.db_tags_name_index:
                version = vistrail.db_tags_name_index[version].db_id

    persistent_module_ids = set()
    for action in vistrail.db_actions:
        for op in action.db_operations:
            if op.db_what == 'module' and (op.vtType == 'add' or 
                                           op.vtType == 'change'):
                module = op.db_data
                if module.db_package in persistence_pkg_ids:
                    persistent_module_ids.add(module.db_id)
                
    filenames = {}
    tags = {}
    for workflow_exec in log.db_workflow_execs:
        cur_version = workflow_exec.db_parent_version
        if version is not None and cur_version != version:
            continue
        if cur_version in vistrail.db_tags_id_index:
            tags[cur_version] = vistrail.db_tags_id_index[cur_version].db_name
        
        for module_exec in workflow_exec.db_item_execs:
            if module_exec.db_module_id in persistent_module_ids:
                for annotation in module_exec.db_annotations:
                    if annotation.db_key == 'signature':
                        if cur_version not in filenames:
                            filenames[cur_version] = set()
                        val = annotation.db_value.upper()
                        filenames[cur_version].add(os.path.join(val[:2], 
                                                                val[2:]))
    shutil.rmtree(save_dir)
    return filenames, tags
                        
if __name__ == '__main__':
    fname = sys.argv[1]
    version = sys.argv[2] if len(sys.argv) > 2 else None
    if version:
        try:
            version = int(version)
        except ValueError:
            pass
    all_files, tags = find_files(fname, version)
    for version, filenames in all_files.iteritems():
        if version in tags:
            print str(version) + ' (' + tags[version] + '):'
        else:
            print str(version) + ':'
        for fname in filenames:
            print ' ', fname
                    



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore, QtGui

from vistrails.core.collection import Collection
from vistrails.core.collection.vistrail import VistrailEntity
from vistrails.core.data_structures.bijectivedict import Bidict
from vistrails.core.query.combined import CombinedSearch
from vistrails.core.query.multiple import MultipleSearch
from vistrails.core.query.version import SearchCompiler, SearchParseError, TrueSearch
from vistrails.core.query.visual import VisualQuery
from vistrails.core.vistrail.pipeline import Pipeline
from vistrails.core.vistrail.vistrail import Vistrail

from vistrails.gui.base_view import BaseView
from vistrails.gui.common_widgets import QSearchBox
from vistrails.gui.modules.utils import get_query_widget_class
from vistrails.gui.pipeline_view import QPipelineView
from vistrails.gui.ports_pane import ParameterEntry
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.version_view import QVersionTreeView
from vistrails.gui.vistrail_controller import VistrailController

class QueryController(object):
    LEVEL_ALL = 0
    LEVEL_VISTRAIL = 1
    LEVEL_WORKFLOW = 2

    def __init__(self, query_view=None):
        self.query_view = query_view
        self.search = None
        self.search_str = None
        self.search_pipeline = None
        self.search_level = 3
        self.use_regex = False
        self.vt_controller = None
        self.level = QueryController.LEVEL_VISTRAIL
        self.workflow_version = None

    def set_level(self, level):
        self.query_view.query_box.setLevel(level)
        self.level_changed(level)

    def set_use_regex(self, use_regex):
        self.use_regex = use_regex

    def level_changed(self, level):
        self.query_view.set_result_level(level)
        if level >= QueryController.LEVEL_VISTRAIL and \
                self.query_view.query_box.backButton.isEnabled():
            self.query_view.query_box.editButton.setEnabled(True)
        else:
            self.query_view.query_box.editButton.setEnabled(False)
        self.level = level

    def set_query_view(self, query_view=None):
        self.query_view = query_view
    
    def set_vistrail_controller(self, vt_controller):
        self.vt_controller = vt_controller

    def set_search(self, search=None):
        self.search = search
        self.query_view.version_result_view.controller.search = search
        self.query_view.workflow_result_view.controller.search = search
        
    def run_search(self, search_str=None):
        """ set_search(search_str: str) -> None
        Change the currrent version tree search statement
        
        """
        search_pipeline = \
            self.query_view.pipeline_view.scene().current_pipeline
        if search_str is None:
            search_str = self.query_view.query_box.getCurrentText()
        self.query_view.update_controller()
        if self.search is None or \
                self.search.search_str != search_str or \
                self.search.queryPipeline != search_pipeline or \
                self.query_view.p_controller.changed or \
                self.search_level > self.level:
            self.search_str = search_str
            self.search_pipeline = search_pipeline
            self.search_level = self.level
            # reset changed here
            self.query_view.p_controller.set_changed(False)
            vt_controller = self.query_view.vt_controller

            def do_search(only_current_vistrail=False, 
                          only_current_workflow=False):
                entities_to_check = {}
                open_col = Collection.getInstance()
                
                for entity in open_col.get_current_entities():
                    if entity.type_id == VistrailEntity.type_id and \
                            entity.is_open:
                        controller = entity._window.controller
                        if only_current_vistrail and \
                                controller.vistrail != vt_controller.vistrail:
                            continue
                        if only_current_workflow:
                            versions_to_check = controller.current_version
                        else:
                            graph = controller._current_terse_graph
                            versions_to_check = set(graph.vertices.iterkeys())
                        entities_to_check[entity] = versions_to_check
                self.set_search(MultipleSearch(search_str, search_pipeline,
                                               entities_to_check,
                                               self.use_regex))
                self.search.run()
                return self.search.getResultEntities()
                
            if self.level == QueryController.LEVEL_VISTRAIL:
                result_entities = do_search(True)                
                self.show_vistrail_matches()
            elif self.level == QueryController.LEVEL_WORKFLOW:
                self.search_level = QueryController.LEVEL_VISTRAIL
                result_entities = do_search(True)
                self.update_version_tree()
                self.show_workflow_matches()
            else:  # self.level == QueryController.LEVEL_ALL
                result_entities = do_search()
                self.show_global_matches()

            from vistrails.gui.vistrails_window import _app
            _app.notify("search_changed", self.search, result_entities)
        else:
            self.query_view.set_to_result_mode()
            
    def set_refine(self, refine):
        """ set_refine(refine: bool) -> None
        Set the refine state to True or False
        
        """
        self.query_view.version_result_view.controller.set_refine(refine)

    def reset_search(self):
        self.search = None
        self.search_pipeline = None
        self.query_view.pipeline_view.controller.change_selected_version(0)
        self.query_view.pipeline_view.scene().setupScene(
            self.query_view.pipeline_view.controller.current_pipeline)
        self.query_view.set_to_search_mode()
        self.query_view.query_box.searchBox.clearSearch()
        self.query_view.vistrailChanged()

        from vistrails.gui.vistrails_window import _app
        _app.notify("search_changed", None, None)

    def back_to_search(self):
        self.query_view.set_to_search_mode()

    def goto_edit(self):
        # get the version info and send it to open_vistrail call
        if self.level == QueryController.LEVEL_VISTRAIL:
            from vistrails.gui.vistrails_window import _app
            _app.qactions['history'].trigger()
        elif self.level == QueryController.LEVEL_WORKFLOW:
            from vistrails.gui.vistrails_window import _app
            version = self.query_view.vt_controller.current_version
            self.query_view.controller.change_selected_version(version, True)
            _app.qactions['pipeline'].trigger()

    def update_results(self):
        if self.workflow_version != \
                self.query_view.vt_controller.current_version:
            result_view = self.query_view.workflow_result_view
            result_view.scene().setupScene(
                result_view.controller.current_pipeline)
            result_view.scene().fitToView(result_view, True)
            self.workflow_version = \
                self.query_view.vt_controller.current_version

    def update_version_tree(self):
        result_view = self.query_view.version_result_view
        if result_view.controller.refine:
            result_view.controller.recompute_terse_graph()
        result_view.controller.invalidate_version_tree()

    def show_vistrail_matches(self, *args, **kwargs):
        if self.level != QueryController.LEVEL_VISTRAIL:
            self.set_level(QueryController.LEVEL_VISTRAIL)
        self.query_view.set_to_result_mode()
        result_view = self.query_view.version_result_view
        if result_view.controller.refine:
            result_view.controller.recompute_terse_graph()
        result_view.controller.invalidate_version_tree(*args, **kwargs)        

    def show_workflow_matches(self):
        if self.level != QueryController.LEVEL_WORKFLOW:
            self.set_level(QueryController.LEVEL_WORKFLOW)
        self.query_view.set_to_result_mode()
        result_view = self.query_view.workflow_result_view
        result_view.scene().setupScene(result_view.controller.current_pipeline)
        result_view.scene().fitToView(result_view, True)

    def show_global_matches(self):
        if self.level != QueryController.LEVEL_ALL:
            self.set_level(QueryController.LEVEL_ALL)
        self.query_view.set_to_result_mode()        

    # def invalidate_version_tree(self, *args, **kwargs):
    #     self.query_view.set_to_result_mode()
    #     result_view = self.query_view.version_result_view
    #     result_view.controller.search = self.search
    #     result_view.controller.search_str = self.search_str
    #     result_view.controller.invalidate_version_tree(*args, **kwargs)

    # def recompute_terse_graph(self, *args, **kwargs):
    #     self.query_view.set_to_result_mode()
    #     result_view = self.query_view.version_result_view
    #     result_view.controller.search = self.search
    #     result_view.controller.search_str = self.search_str
    #     result_view.controller.recompute_terse_graph(*args, **kwargs)


class QQueryPipelineView(QPipelineView):
    def __init__(self, parent=None):
        QPipelineView.__init__(self, parent)
        self.setBackgroundBrush(CurrentTheme.QUERY_BACKGROUND_BRUSH)
        self.scene().current_pipeline = Pipeline()
        self.query_controller = None
      
    def set_query_controller(self, controller):
        self.query_controller = controller
  
    def execute(self):
        self.query_controller.run_search(None)
    
class QQueryResultGlobalView(QtGui.QWidget):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        layout = QtGui.QVBoxLayout()
        label = QtGui.QLabel("See Workspace Window")
        label.setAlignment(QtCore.Qt.AlignCenter)
        layout.addWidget(label, QtCore.Qt.AlignCenter)
        self.setLayout(layout)
        # self.setBackgroundBrush(CurrentTheme.QUERY_RESULT_BACKGROUND_BRUSH)

class QQueryResultVersionView(QVersionTreeView):
    def __init__(self, parent=None):
        QVersionTreeView.__init__(self, parent)
        self.setBackgroundBrush(CurrentTheme.QUERY_RESULT_BACKGROUND_BRUSH)

class QQueryResultWorkflowView(QPipelineView):
    def __init__(self, parent=None):
        QPipelineView.__init__(self, parent)
        self.setBackgroundBrush(CurrentTheme.QUERY_RESULT_BACKGROUND_BRUSH)
        self.scene().set_read_only_mode(True)
    
class QQueryBox(QtGui.QWidget):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.build_widget()
        self.controller = None

    def set_controller(self, controller=None):
        self.controller = controller

    def build_widget(self):
        layout = QtGui.QVBoxLayout()
        layout.setMargin(4)
        layout.setSpacing(2)
        self.searchBox = QSearchBox(True, False, self)
        layout.addWidget(self.searchBox)
        options_layout = QtGui.QHBoxLayout()
        options_layout.setSpacing(5)
        options_layout.setAlignment(QtCore.Qt.AlignLeft)
        options_layout.addWidget(QtGui.QLabel("Search:"))
        searchAll = QtGui.QRadioButton("Open Vistrails")
        searchCurrent = QtGui.QRadioButton("Current Vistrail")
        searchWorkflow = QtGui.QRadioButton("Current Workflow")
        useRegex = QtGui.QCheckBox("Regular expression")
        self.level_group = QtGui.QButtonGroup()
        self.level_group.addButton(searchAll)
        self.level_group.addButton(searchCurrent)
        self.level_group.addButton(searchWorkflow)
        self.level_map = \
            Bidict([(QueryController.LEVEL_ALL, searchAll),
                    (QueryController.LEVEL_VISTRAIL, searchCurrent),
                    (QueryController.LEVEL_WORKFLOW, searchWorkflow)])
        options_layout.addWidget(searchAll)
        options_layout.addWidget(searchCurrent)
        options_layout.addWidget(searchWorkflow)
        options_layout.addWidget(useRegex)
        searchCurrent.setChecked(True)
        
        self.editButton = QtGui.QPushButton("Edit")
        self.editButton.setEnabled(False)
        self.backButton = QtGui.QPushButton("Back to Search")
        self.backButton.setEnabled(False)
        options_layout.addStretch(1)
        options_layout.addWidget(self.editButton, 0, QtCore.Qt.AlignRight)
        options_layout.addWidget(self.backButton, 0, QtCore.Qt.AlignRight)
        layout.addLayout(options_layout)
        self.setLayout(layout)

        self.connect(self.searchBox, QtCore.SIGNAL('resetSearch()'),
                     self.resetSearch)
        self.connect(self.searchBox, QtCore.SIGNAL('executeSearch(QString)'),
                     self.executeSearch)
        self.connect(self.searchBox, QtCore.SIGNAL('refineMode(bool)'),
                     self.refineMode)
        self.connect(self.backButton, QtCore.SIGNAL('clicked()'),
                     self.backToSearch)
        self.connect(self.editButton, QtCore.SIGNAL('clicked()'),
                     self.doEdit)
        self.connect(self.level_group, 
                     QtCore.SIGNAL('buttonClicked(QAbstractButton*)'),
                     self.levelChanged)
        self.connect(useRegex, QtCore.SIGNAL('stateChanged(int)'),
                     self.useRegexChanged)

    def resetSearch(self, emit_signal=True):
        """
        resetSearch() -> None

        """
        if self.controller and emit_signal:
            self.controller.reset_search()
            self.emit(QtCore.SIGNAL('textQueryChange(bool)'), False)
        else:
            self.searchBox.clearSearch()

    def backToSearch(self):
        if self.controller:
            self.controller.back_to_search()

    def doEdit(self):
        if self.controller:
            self.controller.goto_edit()

    def levelChanged(self, button):
        self.controller.set_level(self.level_map.inverse[button])

    def useRegexChanged(self, status):
        self.controller.set_use_regex(status != QtCore.Qt.Unchecked)

    def setLevel(self, level):
        self.level_map[level].setChecked(True)

    def executeSearch(self, text):
        """
        executeSearch(text: QString) -> None

        """
        s = str(text)
        if self.controller:
            self.controller.run_search(s)
            # try:
            #     search = CombinedSearch(s, 
            #     search = SearchCompiler(s).searchStmt
            # except SearchParseError, e:
            #     debug.warning("Search Parse Error", e)
            #     search = None
            # self.controller.set_search(search, s)
            # self.emit(QtCore.SIGNAL('textQueryChange(bool)'), s!='')

    def refineMode(self, on):
        """
        refineMode(on: bool) -> None
        
        """
        if self.controller:
            self.controller.set_refine(on)

    def getCurrentText(self):
        return self.searchBox.getCurrentText()

    def setManualResetEnabled(self, boolVal):
        self.searchBox.setManualResetEnabled(boolVal)

class QQueryView(QtGui.QWidget, BaseView):
    VISUAL_SEARCH_VIEW = 0
    GLOBAL_RESULT_VIEW = 1
    VERSION_RESULT_VIEW = 2
    WORKFLOW_RESULT_VIEW = 3

    RESULT_LEVEL_MAP = \
        Bidict([(QueryController.LEVEL_ALL, GLOBAL_RESULT_VIEW),
                (QueryController.LEVEL_VISTRAIL, VERSION_RESULT_VIEW),
                (QueryController.LEVEL_WORKFLOW, WORKFLOW_RESULT_VIEW)])

    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        BaseView.__init__(self)
        self.build_widget()
        self.set_title("Search")

    def set_controller(self, controller=None):
        if self.controller:
            self.disconnect(self.controller,
                     QtCore.SIGNAL('stateChanged'),
                     self.update_controller)
        self.controller = controller
        if controller:
            self.connect(self.controller,
                         QtCore.SIGNAL('stateChanged'),
                         self.update_controller)
        self.vt_controller.vistrail_view = self.version_result_view
        self.vt_controller.current_pipeline_view = \
            self.workflow_result_view
        # self.vt_controller.vistrail_view.set_controller(self.vt_controller)
        # FIXME Need to figure out how to deal with this !!!
        self.vt_controller.set_vistrail(controller.vistrail, None,
                                        set_log_on_vt=False)
        self.vt_controller.change_selected_version(controller.current_version)
        self.version_result_view.set_controller(self.vt_controller)
        self.workflow_result_view.set_controller(self.vt_controller)
        self.query_controller.set_vistrail_controller(controller)

    def update_controller(self):
        # FIXME Need to figure out how to deal with this !!!
        self.vt_controller.set_vistrail(self.controller.vistrail, None,
                                        set_log_on_vt=False)
        self.vt_controller.change_selected_version(
            self.controller.current_version)

    def build_widget(self):
        layout = QtGui.QVBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(0)

        self.query_controller = QueryController(self)
        self.vt_controller = VistrailController(auto_save=False)
        self.p_controller = VistrailController(Vistrail(), auto_save=False)
        
        self.connect(self.p_controller,
                     QtCore.SIGNAL('vistrailChanged()'),
                     self.vistrailChanged)

        self.query_box = QQueryBox()
        self.query_box.set_controller(self.query_controller)
        layout.addWidget(self.query_box)

        self.stacked_widget = QtGui.QStackedWidget()
        self.pipeline_view = QQueryPipelineView()
        self.p_controller.current_pipeline_view = self.pipeline_view
        self.pipeline_view.set_controller(self.p_controller)
        self.pipeline_view.set_query_controller(self.query_controller)
        QQueryView.VISUAL_SEARCH_VIEW = \
            self.stacked_widget.addWidget(self.pipeline_view)
        self.global_result_view = QQueryResultGlobalView()
        QQueryView.GLOBAL_RESULT_VIEW = \
            self.stacked_widget.addWidget(self.global_result_view)
        self.version_result_view = QQueryResultVersionView()
        self.connect(self.version_result_view.scene(), 
                     QtCore.SIGNAL('versionSelected(int,bool,bool,bool,bool)'),
                     self.result_version_selected)
        # self.version_result_view.set_controller(self.vt_controller)
        QQueryView.VERSION_RESULT_VIEW = \
            self.stacked_widget.addWidget(self.version_result_view)
        self.workflow_result_view = QQueryResultWorkflowView()
        # self.workflow_result_view.set_controller(self.vt_controller)
        QQueryView.WORKFLOW_RESULT_VIEW = \
            self.stacked_widget.addWidget(self.workflow_result_view)
        self.stacked_widget.setCurrentWidget(self.pipeline_view)
        layout.addWidget(self.stacked_widget)

        self.setLayout(layout)
        self.current_display = QQueryView.VISUAL_SEARCH_VIEW
        self.current_result_view = QQueryView.VERSION_RESULT_VIEW

    def set_default_layout(self):
        from vistrails.gui.module_palette import QModulePalette
        from vistrails.gui.module_info import QModuleInfo
        self.set_palette_layout(
            {QtCore.Qt.LeftDockWidgetArea: QModulePalette,
             QtCore.Qt.RightDockWidgetArea: QModuleInfo,
             })
            
    def set_action_links(self):
        self.action_links = \
            { 'execute': ('query_pipeline_changed', self.set_execute_action) }

        # also add other notification here...
        from vistrails.gui.vistrails_window import _app
        _app.register_notification('query_pipeline_changed', 
                                   self.set_reset_button)

    def set_reset_button(self, pipeline):
        self.query_box.setManualResetEnabled(self.pipeline_non_empty(pipeline))

    def set_result_level(self, level):
        view_idx = QQueryView.RESULT_LEVEL_MAP[level]
        if self.current_display != QQueryView.VISUAL_SEARCH_VIEW:
            self.set_display_view(view_idx)
        self.current_result_view = view_idx
        self.query_controller.update_results()
            
    def set_to_search_mode(self):
        self.set_display_view(QQueryView.VISUAL_SEARCH_VIEW)
        self.query_box.backButton.setEnabled(False)
        self.query_box.editButton.setEnabled(False)
        self.set_reset_button(self.p_controller.current_pipeline)

        from vistrails.gui.vistrails_window import _app
        _app.notify('query_pipeline_changed', 
                    self.p_controller.current_pipeline)

    def set_to_result_mode(self):
        self.set_display_view(self.current_result_view)
        self.query_box.backButton.setEnabled(True)
        if self.query_controller.level >= QueryController.LEVEL_VISTRAIL:
            self.query_box.editButton.setEnabled(True)
        self.query_box.setManualResetEnabled(True)

        from vistrails.gui.vistrails_window import _app
        _app.notify('query_pipeline_changed', 
                    self.p_controller.current_pipeline)

    def set_display_view(self, view_type):
        self.current_display = view_type
        self.stacked_widget.setCurrentIndex(view_type)

    def get_current_view(self):
        return self.stacked_widget.currentWidget()
    
    def set_action_defaults(self):
        self.action_defaults = \
            {
             'execute': [('setEnabled', True, self.set_execute_action),
                          ('setIcon', False, CurrentTheme.VISUAL_QUERY_ICON),
                          ('setToolTip', False, 'Execute a visual query')],
             'publishWeb': [('setEnabled', False, False)],
             'publishPaper': [('setEnabled', False, False)],
            }
    
    def set_execute_action(self, pipeline=None):
        if not self.vt_controller:
            return False
        if pipeline is None:
            pipeline = self.p_controller.current_pipeline            
        if self.current_display == QQueryView.VISUAL_SEARCH_VIEW:
            return self.pipeline_non_empty(pipeline)
        return False

    def pipeline_non_empty(self, pipeline):
        return pipeline is not None and len(pipeline.modules) > 0
    
    def vistrailChanged(self):
        from vistrails.gui.vistrails_window import _app
        self.p_controller.current_pipeline.ensure_connection_specs()
        _app.notify('query_pipeline_changed', self.p_controller.current_pipeline)

    def query_changed(self, query=None):
        if query is None:
            self.query_controller.reset_search()
        # FIXME add support for changing the query to something specific

    # DAK: removed this call as the query view maintains its own
    # "current version"
    # def version_changed(self, version_id):
    #     self.vt_controller.change_selected_version(version_id)
    #     self.version_result_view.select_current_version()
    #     self.query_controller.update_results()
        
    def result_version_selected(self, version_id, by_click, do_validate=True,
                                from_root=False, double_click=False):
        if by_click:
            self.query_controller.search.setCurrentVistrail(
                self.vt_controller.vistrail)
            self.vt_controller.change_selected_version(version_id, by_click, 
                                                       do_validate, from_root)
            if double_click:
                self.query_controller.set_level(QueryController.LEVEL_WORKFLOW)
                self.query_controller.show_workflow_matches()

class QueryEntry(ParameterEntry):
    def __init__(self, port_spec, function=None, parent=None):
        ParameterEntry.__init__(self, port_spec, function, parent)

    def get_widget(self):
        return self.build_widget(get_query_widget_class, False)
